"""This file implements the gym environment of minitaur.

"""
import math
import time

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import pybullet
from pybullet_utils import bullet_client as bc
import pybullet_data
from pybullet_envs.minitaur.envs import minitaur
from pybullet_envs.minitaur.envs import minitaur_derpy
from pybullet_envs.minitaur.envs import minitaur_logging
from pybullet_envs.minitaur.envs import minitaur_logging_pb2
from pybullet_envs.minitaur.envs import minitaur_rainbow_dash
from pybullet_envs.minitaur.envs import motor
from pkg_resources import parse_version

NUM_MOTORS = 8
MOTOR_ANGLE_OBSERVATION_INDEX = 0
MOTOR_VELOCITY_OBSERVATION_INDEX = MOTOR_ANGLE_OBSERVATION_INDEX + NUM_MOTORS
MOTOR_TORQUE_OBSERVATION_INDEX = MOTOR_VELOCITY_OBSERVATION_INDEX + NUM_MOTORS
BASE_ORIENTATION_OBSERVATION_INDEX = MOTOR_TORQUE_OBSERVATION_INDEX + NUM_MOTORS
ACTION_EPS = 0.01
OBSERVATION_EPS = 0.01
RENDER_HEIGHT = 360
RENDER_WIDTH = 480
SENSOR_NOISE_STDDEV = minitaur.SENSOR_NOISE_STDDEV
DEFAULT_URDF_VERSION = "default"
DERPY_V0_URDF_VERSION = "derpy_v0"
RAINBOW_DASH_V0_URDF_VERSION = "rainbow_dash_v0"
NUM_SIMULATION_ITERATION_STEPS = 300

MINIATUR_URDF_VERSION_MAP = {
    DEFAULT_URDF_VERSION: minitaur.Minitaur,
    DERPY_V0_URDF_VERSION: minitaur_derpy.MinitaurDerpy,
    RAINBOW_DASH_V0_URDF_VERSION: minitaur_rainbow_dash.MinitaurRainbowDash,
}


def convert_to_list(obj):
  try:
    iter(obj)
    return obj
  except TypeError:
    return [obj]


class MinitaurGymEnv(gym.Env):
  """The gym environment for the minitaur.

  It simulates the locomotion of a minitaur, a quadruped robot. The state space
  include the angles, velocities and torques for all the motors and the action
  space is the desired motor angle for each motor. The reward function is based
  on how far the minitaur walks in 1000 steps and penalizes the energy
  expenditure.

  """
  metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 100}

  def __init__(self,
               urdf_root=pybullet_data.getDataPath(),
               urdf_version=None,
               distance_weight=1.0,
               energy_weight=0.005,
               shake_weight=0.0,
               drift_weight=0.0,
               distance_limit=float("inf"),
               observation_noise_stdev=SENSOR_NOISE_STDDEV,
               self_collision_enabled=True,
               motor_velocity_limit=np.inf,
               pd_control_enabled=False,
               leg_model_enabled=True,
               accurate_motor_model_enabled=False,
               remove_default_joint_damping=False,
               motor_kp=1.0,
               motor_kd=0.02,
               control_latency=0.0,
               pd_latency=0.0,
               torque_control_enabled=False,
               motor_overheat_protection=False,
               hard_reset=True,
               on_rack=False,
               render=False,
               num_steps_to_log=1000,
               action_repeat=1,
               control_time_step=None,
               env_randomizer=None,
               forward_reward_cap=float("inf"),
               reflection=True,
               log_path=None):
    """Initialize the minitaur gym environment.

    Args:
      urdf_root: The path to the urdf data folder.
      urdf_version: [DEFAULT_URDF_VERSION, DERPY_V0_URDF_VERSION,
        RAINBOW_DASH_V0_URDF_VERSION] are allowable
        versions. If None, DEFAULT_URDF_VERSION is used. DERPY_V0_URDF_VERSION
        is the result of first pass system identification for derpy.
        We will have a different URDF and related Minitaur class each time we
        perform system identification. While the majority of the code of the
        class remains the same, some code changes (e.g. the constraint location
        might change). __init__() will choose the right Minitaur class from
        different minitaur modules based on
        urdf_version.
      distance_weight: The weight of the distance term in the reward.
      energy_weight: The weight of the energy term in the reward.
      shake_weight: The weight of the vertical shakiness term in the reward.
      drift_weight: The weight of the sideways drift term in the reward.
      distance_limit: The maximum distance to terminate the episode.
      observation_noise_stdev: The standard deviation of observation noise.
      self_collision_enabled: Whether to enable self collision in the sim.
      motor_velocity_limit: The velocity limit of each motor.
      pd_control_enabled: Whether to use PD controller for each motor.
      leg_model_enabled: Whether to use a leg motor to reparameterize the action
        space.
      accurate_motor_model_enabled: Whether to use the accurate DC motor model.
      remove_default_joint_damping: Whether to remove the default joint damping.
      motor_kp: proportional gain for the accurate motor model.
      motor_kd: derivative gain for the accurate motor model.
      control_latency: It is the delay in the controller between when an
        observation is made at some point, and when that reading is reported
        back to the Neural Network.
      pd_latency: latency of the PD controller loop. PD calculates PWM based on
        the motor angle and velocity. The latency measures the time between when
        the motor angle and velocity are observed on the microcontroller and
        when the true state happens on the motor. It is typically (0.001-
        0.002s).
      torque_control_enabled: Whether to use the torque control, if set to
        False, pose control will be used.
      motor_overheat_protection: Whether to shutdown the motor that has exerted
        large torque (OVERHEAT_SHUTDOWN_TORQUE) for an extended amount of time
        (OVERHEAT_SHUTDOWN_TIME). See ApplyAction() in minitaur.py for more
        details.
      hard_reset: Whether to wipe the simulation and load everything when reset
        is called. If set to false, reset just place the minitaur back to start
        position and set its pose to initial configuration.
      on_rack: Whether to place the minitaur on rack. This is only used to debug
        the walking gait. In this mode, the minitaur's base is hanged midair so
        that its walking gait is clearer to visualize.
      render: Whether to render the simulation.
      num_steps_to_log: The max number of control steps in one episode that will
        be logged. If the number of steps is more than num_steps_to_log, the
        environment will still be running, but only first num_steps_to_log will
        be recorded in logging.
      action_repeat: The number of simulation steps before actions are applied.
      control_time_step: The time step between two successive control signals.
      env_randomizer: An instance (or a list) of EnvRandomizer(s). An
        EnvRandomizer may randomize the physical property of minitaur, change
          the terrrain during reset(), or add perturbation forces during step().
      forward_reward_cap: The maximum value that forward reward is capped at.
        Disabled (Inf) by default.
      log_path: The path to write out logs. For the details of logging, refer to
        minitaur_logging.proto.
    Raises:
      ValueError: If the urdf_version is not supported.
    """
    # Set up logging.
    self._log_path = log_path
    self.logging = minitaur_logging.MinitaurLogging(log_path)
    # PD control needs smaller time step for stability.
    if control_time_step is not None:
      self.control_time_step = control_time_step
      self._action_repeat = action_repeat
      self._time_step = control_time_step / action_repeat
    else:
      # Default values for time step and action repeat
      if accurate_motor_model_enabled or pd_control_enabled:
        self._time_step = 0.002
        self._action_repeat = 5
      else:
        self._time_step = 0.01
        self._action_repeat = 1
      self.control_time_step = self._time_step * self._action_repeat
    # TODO(b/73829334): Fix the value of self._num_bullet_solver_iterations.
    self._num_bullet_solver_iterations = int(NUM_SIMULATION_ITERATION_STEPS / self._action_repeat)
    self._urdf_root = urdf_root
    self._self_collision_enabled = self_collision_enabled
    self._motor_velocity_limit = motor_velocity_limit
    self._observation = []
    self._true_observation = []
    self._objectives = []
    self._objective_weights = [distance_weight, energy_weight, drift_weight, shake_weight]
    self._env_step_counter = 0
    self._num_steps_to_log = num_steps_to_log
    self._is_render = render
    self._last_base_position = [0, 0, 0]
    self._distance_weight = distance_weight
    self._energy_weight = energy_weight
    self._drift_weight = drift_weight
    self._shake_weight = shake_weight
    self._distance_limit = distance_limit
    self._observation_noise_stdev = observation_noise_stdev
    self._action_bound = 1
    self._pd_control_enabled = pd_control_enabled
    self._leg_model_enabled = leg_model_enabled
    self._accurate_motor_model_enabled = accurate_motor_model_enabled
    self._remove_default_joint_damping = remove_default_joint_damping
    self._motor_kp = motor_kp
    self._motor_kd = motor_kd
    self._torque_control_enabled = torque_control_enabled
    self._motor_overheat_protection = motor_overheat_protection
    self._on_rack = on_rack
    self._cam_dist = 1.0
    self._cam_yaw = 0
    self._cam_pitch = -30
    self._forward_reward_cap = forward_reward_cap
    self._hard_reset = True
    self._last_frame_time = 0.0
    self._control_latency = control_latency
    self._pd_latency = pd_latency
    self._urdf_version = urdf_version
    self._ground_id = None
    self._reflection = reflection
    self._env_randomizers = convert_to_list(env_randomizer) if env_randomizer else []
    self._episode_proto = minitaur_logging_pb2.MinitaurEpisode()
    if self._is_render:
      self._pybullet_client = bc.BulletClient(connection_mode=pybullet.GUI)
    else:
      self._pybullet_client = bc.BulletClient()
    if self._urdf_version is None:
      self._urdf_version = DEFAULT_URDF_VERSION
    self._pybullet_client.setPhysicsEngineParameter(enableConeFriction=0)
    self.seed()
    self.reset()
    observation_high = (self._get_observation_upper_bound() + OBSERVATION_EPS)
    observation_low = (self._get_observation_lower_bound() - OBSERVATION_EPS)
    action_dim = NUM_MOTORS
    action_high = np.array([self._action_bound] * action_dim)
    self.action_space = spaces.Box(-action_high, action_high)
    self.observation_space = spaces.Box(observation_low, observation_high)
    self.viewer = None
    self._hard_reset = hard_reset  # This assignment need to be after reset()

  def close(self):
    if self._env_step_counter > 0:
      self.logging.save_episode(self._episode_proto)
    self.minitaur.Terminate()

  def add_env_randomizer(self, env_randomizer):
    self._env_randomizers.append(env_randomizer)

  def reset(self, initial_motor_angles=None, reset_duration=1.0):
    self._pybullet_client.configureDebugVisualizer(self._pybullet_client.COV_ENABLE_RENDERING, 0)
    if self._env_step_counter > 0:
      self.logging.save_episode(self._episode_proto)
    self._episode_proto = minitaur_logging_pb2.MinitaurEpisode()
    minitaur_logging.preallocate_episode_proto(self._episode_proto, self._num_steps_to_log)
    if self._hard_reset:
      self._pybullet_client.resetSimulation()
      self._pybullet_client.setPhysicsEngineParameter(
          numSolverIterations=int(self._num_bullet_solver_iterations))
      self._pybullet_client.setTimeStep(self._time_step)
      self._ground_id = self._pybullet_client.loadURDF("%s/plane.urdf" % self._urdf_root)
      if (self._reflection):
        self._pybullet_client.changeVisualShape(self._ground_id, -1, rgbaColor=[1, 1, 1, 0.8])
        self._pybullet_client.configureDebugVisualizer(
            self._pybullet_client.COV_ENABLE_PLANAR_REFLECTION, self._ground_id)
      self._pybullet_client.setGravity(0, 0, -10)
      acc_motor = self._accurate_motor_model_enabled
      motor_protect = self._motor_overheat_protection
      if self._urdf_version not in MINIATUR_URDF_VERSION_MAP:
        raise ValueError("%s is not a supported urdf_version." % self._urdf_version)
      else:
        self.minitaur = (MINIATUR_URDF_VERSION_MAP[self._urdf_version](
            pybullet_client=self._pybullet_client,
            action_repeat=self._action_repeat,
            urdf_root=self._urdf_root,
            time_step=self._time_step,
            self_collision_enabled=self._self_collision_enabled,
            motor_velocity_limit=self._motor_velocity_limit,
            pd_control_enabled=self._pd_control_enabled,
            accurate_motor_model_enabled=acc_motor,
            remove_default_joint_damping=self._remove_default_joint_damping,
            motor_kp=self._motor_kp,
            motor_kd=self._motor_kd,
            control_latency=self._control_latency,
            pd_latency=self._pd_latency,
            observation_noise_stdev=self._observation_noise_stdev,
            torque_control_enabled=self._torque_control_enabled,
            motor_overheat_protection=motor_protect,
            on_rack=self._on_rack))
    self.minitaur.Reset(reload_urdf=False,
                        default_motor_angles=initial_motor_angles,
                        reset_time=reset_duration)

    # Loop over all env randomizers.
    for env_randomizer in self._env_randomizers:
      env_randomizer.randomize_env(self)

    self._pybullet_client.setPhysicsEngineParameter(enableConeFriction=0)
    self._env_step_counter = 0
    self._last_base_position = [0, 0, 0]
    self._objectives = []
    self._pybullet_client.resetDebugVisualizerCamera(self._cam_dist, self._cam_yaw,
                                                     self._cam_pitch, [0, 0, 0])
    self._pybullet_client.configureDebugVisualizer(self._pybullet_client.COV_ENABLE_RENDERING, 1)
    return self._get_observation()

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def _transform_action_to_motor_command(self, action):
    if self._leg_model_enabled:
      for i, action_component in enumerate(action):
        if not (-self._action_bound - ACTION_EPS <= action_component <=
                self._action_bound + ACTION_EPS):
          raise ValueError("{}th action {} out of bounds.".format(i, action_component))
      action = self.minitaur.ConvertFromLegModel(action)
    return action

  def step(self, action):
    """Step forward the simulation, given the action.

    Args:
      action: A list of desired motor angles for eight motors.

    Returns:
      observations: The angles, velocities and torques of all motors.
      reward: The reward for the current state-action pair.
      done: Whether the episode has ended.
      info: A dictionary that stores diagnostic information.

    Raises:
      ValueError: The action dimension is not the same as the number of motors.
      ValueError: The magnitude of actions is out of bounds.
    """
    self._last_base_position = self.minitaur.GetBasePosition()

    if self._is_render:
      # Sleep, otherwise the computation takes less time than real time,
      # which will make the visualization like a fast-forward video.
      time_spent = time.time() - self._last_frame_time
      self._last_frame_time = time.time()
      time_to_sleep = self.control_time_step - time_spent
      if time_to_sleep > 0:
        time.sleep(time_to_sleep)
      base_pos = self.minitaur.GetBasePosition()
      # Keep the previous orientation of the camera set by the user.
      [yaw, pitch, dist] = self._pybullet_client.getDebugVisualizerCamera()[8:11]
      self._pybullet_client.resetDebugVisualizerCamera(dist, yaw, pitch, base_pos)

    for env_randomizer in self._env_randomizers:
      env_randomizer.randomize_step(self)

    action = self._transform_action_to_motor_command(action)
    self.minitaur.Step(action)
    reward = self._reward()
    done = self._termination()
    if self._log_path is not None:
      minitaur_logging.update_episode_proto(self._episode_proto, self.minitaur, action,
                                            self._env_step_counter)
    self._env_step_counter += 1
    if done:
      self.minitaur.Terminate()
    return np.array(self._get_observation()), reward, done, {}

  def render(self, mode="rgb_array", close=False):
    if mode != "rgb_array":
      return np.array([])
    base_pos = self.minitaur.GetBasePosition()
    view_matrix = self._pybullet_client.computeViewMatrixFromYawPitchRoll(
        cameraTargetPosition=base_pos,
        distance=self._cam_dist,
        yaw=self._cam_yaw,
        pitch=self._cam_pitch,
        roll=0,
        upAxisIndex=2)
    proj_matrix = self._pybullet_client.computeProjectionMatrixFOV(fov=60,
                                                                   aspect=float(RENDER_WIDTH) /
                                                                   RENDER_HEIGHT,
                                                                   nearVal=0.1,
                                                                   farVal=100.0)
    (_, _, px, _, _) = self._pybullet_client.getCameraImage(
        width=RENDER_WIDTH,
        height=RENDER_HEIGHT,
        renderer=self._pybullet_client.ER_BULLET_HARDWARE_OPENGL,
        viewMatrix=view_matrix,
        projectionMatrix=proj_matrix)
    rgb_array = np.array(px)
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def get_minitaur_motor_angles(self):
    """Get the minitaur's motor angles.

    Returns:
      A numpy array of motor angles.
    """
    return np.array(self._observation[MOTOR_ANGLE_OBSERVATION_INDEX:MOTOR_ANGLE_OBSERVATION_INDEX +
                                      NUM_MOTORS])

  def get_minitaur_motor_velocities(self):
    """Get the minitaur's motor velocities.

    Returns:
      A numpy array of motor velocities.
    """
    return np.array(
        self._observation[MOTOR_VELOCITY_OBSERVATION_INDEX:MOTOR_VELOCITY_OBSERVATION_INDEX +
                          NUM_MOTORS])

  def get_minitaur_motor_torques(self):
    """Get the minitaur's motor torques.

    Returns:
      A numpy array of motor torques.
    """
    return np.array(
        self._observation[MOTOR_TORQUE_OBSERVATION_INDEX:MOTOR_TORQUE_OBSERVATION_INDEX +
                          NUM_MOTORS])

  def get_minitaur_base_orientation(self):
    """Get the minitaur's base orientation, represented by a quaternion.

    Returns:
      A numpy array of minitaur's orientation.
    """
    return np.array(self._observation[BASE_ORIENTATION_OBSERVATION_INDEX:])

  def is_fallen(self):
    """Decide whether the minitaur has fallen.

    If the up directions between the base and the world is larger (the dot
    product is smaller than 0.85) or the base is very low on the ground
    (the height is smaller than 0.13 meter), the minitaur is considered fallen.

    Returns:
      Boolean value that indicates whether the minitaur has fallen.
    """
    orientation = self.minitaur.GetBaseOrientation()
    rot_mat = self._pybullet_client.getMatrixFromQuaternion(orientation)
    local_up = rot_mat[6:]
    pos = self.minitaur.GetBasePosition()
    return (np.dot(np.asarray([0, 0, 1]), np.asarray(local_up)) < 0.85 or pos[2] < 0.13)

  def _termination(self):
    position = self.minitaur.GetBasePosition()
    distance = math.sqrt(position[0]**2 + position[1]**2)
    return self.is_fallen() or distance > self._distance_limit

  def _reward(self):
    current_base_position = self.minitaur.GetBasePosition()
    forward_reward = current_base_position[0] - self._last_base_position[0]
    # Cap the forward reward if a cap is set.
    forward_reward = min(forward_reward, self._forward_reward_cap)
    # Penalty for sideways translation.
    drift_reward = -abs(current_base_position[1] - self._last_base_position[1])
    # Penalty for sideways rotation of the body.
    orientation = self.minitaur.GetBaseOrientation()
    rot_matrix = pybullet.getMatrixFromQuaternion(orientation)
    local_up_vec = rot_matrix[6:]
    shake_reward = -abs(np.dot(np.asarray([1, 1, 0]), np.asarray(local_up_vec)))
    energy_reward = -np.abs(
        np.dot(self.minitaur.GetMotorTorques(),
               self.minitaur.GetMotorVelocities())) * self._time_step
    objectives = [forward_reward, energy_reward, drift_reward, shake_reward]
    weighted_objectives = [o * w for o, w in zip(objectives, self._objective_weights)]
    reward = sum(weighted_objectives)
    self._objectives.append(objectives)
    return reward

  def get_objectives(self):
    return self._objectives

  @property
  def objective_weights(self):
    """Accessor for the weights for all the objectives.

    Returns:
      List of floating points that corresponds to weights for the objectives in
      the order that objectives are stored.
    """
    return self._objective_weights

  def _get_observation(self):
    """Get observation of this environment, including noise and latency.

    The minitaur class maintains a history of true observations. Based on the
    latency, this function will find the observation at the right time,
    interpolate if necessary. Then Gaussian noise is added to this observation
    based on self.observation_noise_stdev.

    Returns:
      The noisy observation with latency.
    """

    observation = []
    observation.extend(self.minitaur.GetMotorAngles().tolist())
    observation.extend(self.minitaur.GetMotorVelocities().tolist())
    observation.extend(self.minitaur.GetMotorTorques().tolist())
    observation.extend(list(self.minitaur.GetBaseOrientation()))
    self._observation = observation
    return self._observation

  def _get_true_observation(self):
    """Get the observations of this environment.

    It includes the angles, velocities, torques and the orientation of the base.

    Returns:
      The observation list. observation[0:8] are motor angles. observation[8:16]
      are motor velocities, observation[16:24] are motor torques.
      observation[24:28] is the orientation of the base, in quaternion form.
    """
    observation = []
    observation.extend(self.minitaur.GetTrueMotorAngles().tolist())
    observation.extend(self.minitaur.GetTrueMotorVelocities().tolist())
    observation.extend(self.minitaur.GetTrueMotorTorques().tolist())
    observation.extend(list(self.minitaur.GetTrueBaseOrientation()))

    self._true_observation = observation
    return self._true_observation

  def _get_observation_upper_bound(self):
    """Get the upper bound of the observation.

    Returns:
      The upper bound of an observation. See GetObservation() for the details
        of each element of an observation.
    """
    upper_bound = np.zeros(self._get_observation_dimension())
    num_motors = self.minitaur.num_motors
    upper_bound[0:num_motors] = math.pi  # Joint angle.
    upper_bound[num_motors:2 * num_motors] = (motor.MOTOR_SPEED_LIMIT)  # Joint velocity.
    upper_bound[2 * num_motors:3 * num_motors] = (motor.OBSERVED_TORQUE_LIMIT)  # Joint torque.
    upper_bound[3 * num_motors:] = 1.0  # Quaternion of base orientation.
    return upper_bound

  def _get_observation_lower_bound(self):
    """Get the lower bound of the observation."""
    return -self._get_observation_upper_bound()

  def _get_observation_dimension(self):
    """Get the length of the observation list.

    Returns:
      The length of the observation list.
    """
    return len(self._get_observation())

  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _render = render
    _reset = reset
    _seed = seed
    _step = step

  def set_time_step(self, control_step, simulation_step=0.001):
    """Sets the time step of the environment.

    Args:
      control_step: The time period (in seconds) between two adjacent control
        actions are applied.
      simulation_step: The simulation time step in PyBullet. By default, the
        simulation step is 0.001s, which is a good trade-off between simulation
        speed and accuracy.
    Raises:
      ValueError: If the control step is smaller than the simulation step.
    """
    if control_step < simulation_step:
      raise ValueError("Control step should be larger than or equal to simulation step.")
    self.control_time_step = control_step
    self._time_step = simulation_step
    self._action_repeat = int(round(control_step / simulation_step))
    self._num_bullet_solver_iterations = (NUM_SIMULATION_ITERATION_STEPS / self._action_repeat)
    self._pybullet_client.setPhysicsEngineParameter(
        numSolverIterations=self._num_bullet_solver_iterations)
    self._pybullet_client.setTimeStep(self._time_step)
    self.minitaur.SetTimeSteps(action_repeat=self._action_repeat, simulation_step=self._time_step)

  @property
  def pybullet_client(self):
    return self._pybullet_client

  @property
  def ground_id(self):
    return self._ground_id

  @ground_id.setter
  def ground_id(self, new_ground_id):
    self._ground_id = new_ground_id

  @property
  def env_step_counter(self):
    return self._env_step_counter



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mock environment for testing reinforcement learning code."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gym
import gym.spaces
import numpy as np


class MockEnvironment(object):
  """Generate random agent input and keep track of statistics."""

  def __init__(self, observ_shape, action_shape, min_duration, max_duration):
    """Generate random agent input and keep track of statistics.

    Args:
      observ_shape: Shape for the random observations.
      action_shape: Shape for the action space.
      min_duration: Minimum number of steps per episode.
      max_duration: Maximum number of steps per episode.

    Attributes:
      steps: List of actual simulated lengths for all episodes.
      durations: List of decided lengths for all episodes.
    """
    self._observ_shape = observ_shape
    self._action_shape = action_shape
    self._min_duration = min_duration
    self._max_duration = max_duration
    self._random = np.random.RandomState(0)
    self.steps = []
    self.durations = []

  @property
  def observation_space(self):
    low = np.zeros(self._observ_shape)
    high = np.ones(self._observ_shape)
    return gym.spaces.Box(low, high)

  @property
  def action_space(self):
    low = np.zeros(self._action_shape)
    high = np.ones(self._action_shape)
    return gym.spaces.Box(low, high)

  @property
  def unwrapped(self):
    return self

  def step(self, action):
    assert self.action_space.contains(action)
    assert self.steps[-1] < self.durations[-1]
    self.steps[-1] += 1
    observ = self._current_observation()
    reward = self._current_reward()
    done = self.steps[-1] >= self.durations[-1]
    info = {}
    return observ, reward, done, info

  def reset(self):
    duration = self._random.randint(self._min_duration, self._max_duration + 1)
    self.steps.append(0)
    self.durations.append(duration)
    return self._current_observation()

  def _current_observation(self):
    return self._random.uniform(0, 1, self._observ_shape)

  def _current_reward(self):
    return self._random.uniform(-1, 1)



# Lint as: python3
"""Module for controllers of autonomous objects."""

import abc
import bisect
import enum
from typing import Any, Dict, Optional, Sequence, Text, Tuple, Union

from absl import logging
import dataclasses
import gin
import numpy as np


# A constant to be passed into act as parameter t for initial value.
INIT_TIME = -1.0

# Distance that is deemed close enough in ChaseController.
_EPS_DISTANCE = 1e-4

ControllerOutput = Tuple[np.ndarray, np.ndarray, Dict[Text, Any]]

ANIMATION_FRAME_NUMBER_KEY = "animation_frame_number"


class ControllerBase(metaclass=abc.ABCMeta):
  """Base class of object controllers.

  Controller is similar to a policy in that its output controls autonomous
  object just as policy output controls agent. To reflect this similarity,
  get_action(), the function that "commands" to autonomous object, is named
  similar to the counterpart in policy.
  """

  @abc.abstractmethod
  def get_action(self,
                 time_sec: float,
                 observations: Dict[Text, Any]) -> ControllerOutput:
    """Returns position, orientation and pose based on time and observations.

    Args:
      time_sec: Time since simulation reset in seconds. If time < 0, returns
        initial values and ignores observations.
      observations: A dict of all observations.

    Returns:
      Position, orientation and an extra info dict for robot joints, human
        skeletal pose, etc.
    """


@gin.configurable
class StationaryController(ControllerBase):
  """Controller that keeps constant position and orientation."""

  def __init__(self,
               position: Sequence[float] = None,
               orientation: Sequence[float] = None):
    self._position = np.array(position if position is not None else (0, 0, 0))
    self._orientation = np.array(
        orientation if orientation is not None else (0, 0, 0, 1))

  def get_action(self,
                 t: float,
                 observations: Dict[Text, Any]) -> ControllerOutput:
    """Returns constant position orientation."""
    del t, observations
    return self._position, self._orientation, {}


@gin.configurable
class CircularMotionController(ControllerBase):
  """Controller for circular motion.

  The motion trajectory goes around a center in a circle in xy-plane.
  """

  def __init__(self,
               center: Sequence[float],
               radius: float,
               angular_velocity: float = np.pi,
               face_travel_direction: bool = False):
    """Constructor.

    Args:
      center: Center of circular motion, [x, y, z] in meters.
      radius: Radius of the circle in meters.
      angular_velocity: Angular velocity of motion, unit rad/s, e.g. pi means
        completing a circle in 2 sec.
      face_travel_direction: If True, object will face direction of motion.
    """

    self._center = np.array(center)
    self._radius = radius
    self._angular_velocity = angular_velocity
    self._face_travel_direction = face_travel_direction

  def get_action(self,
                 t: float,
                 observations: Dict[Text, Any]) -> ControllerOutput:
    """Returns position on the circle based on time and constant orientation."""
    del observations

    t = max(0.0, t)
    position = np.array(
        [np.cos(self._angular_velocity * t),
         np.sin(self._angular_velocity * t),
         0]) * self._radius + self._center
    if self._face_travel_direction:
      yaw = self._angular_velocity * t + (
          np.sign(self._angular_velocity) * np.pi / 2)
      orientation = np.array((0, 0, np.sin(yaw / 2), np.cos(yaw / 2)))
    else:
      orientation = np.array((0, 0, 0, 1))
    return position, orientation, {}


class PatrolRepeatMode(enum.Enum):
  """Enums that defines trajectory repeat mode for patrol type controller."""

  # Trajectory does not repeat.  For 3 points a, b, c, the trajectory moves
  # along a -> b -> c and then stops at c forever.
  NO_REPEAT = 0

  # Trajectory repeats as a loop. For 3 points a, b, c, the trajectory moves
  # along a -> b -> c -> a -> b ...
  LOOP = 1

  # Trajectory repeats back tracking previous point first. For 3 points a, b, c,
  # the trajectory moves along a -> b -> c -> b -> a -> b -> c ...
  BACK_TRACK = 2

  # Trajectory repeats by resetting to the initial position after reaching end.
  # For 3 points a, b, c, the trajectory moves a -> b -> c then immediately
  # jumps back to a before continue moving along a -> b -> c again.
  RESET = 3


@dataclasses.dataclass
class PatrolSegmentData:
  """A data class that describes a patrol segment."""

  # Time in a single cycle to start this segment, range [0, cycle_time).
  start_time: float

  # Segment start position.
  start_position: np.ndarray

  # Segment velocity vector.
  velocity: np.ndarray

  # Orientation quaternion of this segment.
  orientation: np.ndarray


@gin.configurable
class WayPointPatrolController(ControllerBase):
  """Controller for patrolling along define waypoints."""

  def __init__(self,
               points: Sequence[Sequence[float]],
               yaw_angle: float = 0,
               face_travel_direction: bool = True,
               repeat_mode: Union[
                   PatrolRepeatMode, Text] = PatrolRepeatMode.NO_REPEAT,
               speed_mps: Optional[float] = 1.0,
               time_points: Optional[Sequence[float]] = None):
    """Constructor.

    Args:
      points: List of waypoints, shape Nx3 or Nx2, N is number of points.
      yaw_angle: Yaw angle of the object in radians.
      face_travel_direction: If True, yaw angle 'zero' will be redefined to be
        object's travel direction. Setting yaw_angle to zero with
        face_travel_direction == True will results in object always facing its
        travel direction. Non-zero yaw_angle will cause additional yaw offsets.
      repeat_mode: Behavior of object after reaching the last way point in list.
        If the value is Text, it is converted to PatrolRepeatMode.
      speed_mps: Speed in meters per second.
      time_points: List of times associated with points. These times
        represent when the object should arrive at the associated waypoint.
        Optional, but if provided it must have the same length as 'points'.
        If 'speed_mps' is None, then 'time_points' will be used as-is and there
        is no maximum segment speed. If 'speed_mps' is also defined, then it
        serves as a maximum speed value and time points which would result in a
        segment speed above this value will be altered such that the maximum
        segment speed is 'speed_mps'.
    """
    self._repeat = (repeat_mode if isinstance(repeat_mode, PatrolRepeatMode)
                    else PatrolRepeatMode[repeat_mode])
    self._yaw_angle = yaw_angle
    self._face_travel_direction = face_travel_direction
    self._speed_mps = speed_mps

    if len(points) < 2:
      raise ValueError(
          f"Need at least two points in 'points', got {len(points)}")

    if time_points is not None and self._repeat is PatrolRepeatMode.LOOP:
      raise ValueError("Time points are not compatible with LOOP mode.")

    if (self._repeat is PatrolRepeatMode.NO_REPEAT or
        self._repeat is PatrolRepeatMode.RESET):
      augmented_points = points

      if time_points is not None:
        augmented_time_points = time_points
    elif self._repeat is PatrolRepeatMode.LOOP:
      augmented_points = list(points) + [points[0]]
    elif self._repeat is PatrolRepeatMode.BACK_TRACK:
      augmented_points = list(points) + list(reversed(points[:-1]))

      if time_points is not None:
        # Time strictly increases, so add the timepoints again on top
        # of the last entry. We add the difference between the last time
        # and the previous elements (in reverse order), on top of the last
        # element where we left off.
        augmented_time_points = list(time_points) + \
            list(time_points[-1] + (time_points[-1] - np.array(time_points[1::-1])))
    else:
      raise NotImplementedError(
          f"Repeat mode {self._repeat} is not supported yet.")

    augmented_points = np.array(augmented_points)
    # For Nx2 inputs, pad it to Nx3 with default z value of 0.
    if augmented_points.shape[1] == 2:
      augmented_points = np.hstack(
          augmented_points, np.zeros(augmented_points.shape[0], 1))
    elif augmented_points.shape[1] != 3:
      raise ValueError("Expect 'points' to be Nx2 or Nx3.")

    t = 0
    self._segments = []

    if time_points is None:
      for from_point, to_point in zip(
          augmented_points[:-1], augmented_points[1:]):
        segment, t = self._get_patrol_segment_by_speed(from_point, to_point, t)
        self._segments.append(segment)
    else:
      for from_point, to_point, from_time, to_time in zip(
          augmented_points[:-1], augmented_points[1:],
          augmented_time_points[:-1], augmented_time_points[1:]):
        segment, t = self._get_patrol_segment_by_time(from_point, to_point, t,
                                                      t + (to_time - from_time))
        self._segments.append(segment)

    self._segment_times = [l.start_time for l in self._segments]

    self._cycle_time = t

  def _get_patrol_segment_by_time(self,
                                  from_point,
                                  to_point,
                                  from_time,
                                  to_time):
    """Returns a PatrolSegmentData for the given points and times."""
    unit_vector, length = self._get_vector(from_point, to_point)
    orientation = self._get_orientation(unit_vector)

    if np.isclose(to_time, from_time):
      speed_mps = 0
    else:
      speed_mps = length / (to_time - from_time)

      if self._speed_mps is not None:
        speed_mps = np.min([self._speed_mps, speed_mps])

    if np.isclose(0, speed_mps):
      new_to_time = to_time
    else:
      new_to_time = np.max([to_time, from_time + (length / speed_mps)])

    segment = PatrolSegmentData(
        start_time=from_time,
        start_position=np.array(from_point),
        velocity=unit_vector * speed_mps,
        orientation=orientation)

    return segment, new_to_time

  def _get_patrol_segment_by_speed(self, from_point, to_point, current_time):
    """Returns a PatrolSegmentData for the given points and a constant speed."""
    unit_vector, length = self._get_vector(from_point, to_point)
    orientation = self._get_orientation(unit_vector)

    segment = PatrolSegmentData(
        start_time=current_time,
        start_position=np.array(from_point),
        velocity=unit_vector * self._speed_mps,
        orientation=orientation)
    time = current_time + length / self._speed_mps

    return segment, time

  def _get_vector(self, from_point, to_point):
    """Gets the unit vector and length of a from/to point pair."""
    vector = np.array(to_point) - np.array(from_point)
    length = np.linalg.norm(vector)

    if length == 0:
      raise ValueError(f"Length of patrol segment equal to 0, "
                       f"from {from_point} to {to_point}.")
    unit_vector = vector / length

    return unit_vector, length

  def _get_orientation(self, unit_vector):
    """Gets the orientation quaternion given a unit vector."""
    yaw = (np.arctan2(unit_vector[1], unit_vector[0])
           if self._face_travel_direction else 0) + self._yaw_angle

    orientation = np.array((0, 0, np.sin(yaw / 2), np.cos(yaw / 2)))

    return orientation

  def get_action(self,
                 t: float,
                 observations: Dict[Text, Any]) -> ControllerOutput:
    """Returns position on, and orientation along the patrol segment."""
    del observations

    # t < 0 means initial condition, which is the same as the value at t = 0.
    t = max(0, t)

    if t > self._cycle_time:
      t = (self._cycle_time if self._repeat is PatrolRepeatMode.NO_REPEAT
           else np.fmod(t, self._cycle_time))

    segment = self._segments[bisect.bisect_right(self._segment_times, t) - 1]
    position = (
        segment.start_position + segment.velocity * (t - segment.start_time))

    return position, segment.orientation.copy(), {}


@gin.configurable
class LinearPatrolController(WayPointPatrolController):
  """Controller for patrolling along a line segment (back and forth)."""

  def __init__(self,
               from_point: Sequence[float],
               to_point: Sequence[float],
               **kwargs):
    """Constructor.

    Args:
      from_point: Starting point of motion, [x, y, z] in meters.
      to_point: Returning point of motion, [x, y, z] in meters.
      **kwargs: Keyword arguments to pass onto base class.
    """
    super().__init__([from_point, to_point],
                     repeat_mode=PatrolRepeatMode.LOOP,
                     **kwargs)


# TODO(b/156126975): migrates this to use difference equation controller.
@gin.configurable
class ChaseController(ControllerBase):
  """Controller for an object to chase another object at certain speed."""

  def __init__(self,
               self_key: Text,
               target_key: Text,
               initial_position: Sequence[float] = (0, 0, 0),
               initial_orientation: Sequence[float] = (0, 0, 0, 1),
               speed_mps: float = 1.0,
               verbose: bool = False):
    """Constructor.

    Args:
      self_key: Observation dict key of position of object being controlled.
      target_key: Observation dict key of position of target object.
      initial_position: Initial position of the object.
      initial_orientation: Initial orientation of the object in xyzw quaternion.
      speed_mps: Speed in meters per second, always positive.
      verbose: If True, log details of get_action() calculation for debugging.
    """
    self._init_position = np.array(initial_position)
    self._init_orientation = np.array(initial_orientation)
    self._previous_orientation = self._init_orientation
    if speed_mps <= 0:
      raise ValueError(
          f"'speed_mps' should be a positive value, got {speed_mps}.")
    self._speed_mps = speed_mps

    self._self_key = self_key
    self._target_key = target_key
    self._verbose = verbose

    self._time_sec = 0

  def get_action(self,
                 time_sec: float,
                 observations: Dict[Text, Any]) -> ControllerOutput:
    """Returns position and orientation of the object being controlled.

    Args:
      time_sec: Time since simulation reset in seconds. If time < 0, returns
        initial values and ignores observations.
      observations: A dict of all observations.
    """
    if time_sec < 0:
      # Initializes internal time.
      self._time_sec = 0
      return self._init_position.copy(), self._init_orientation.copy(), {}

    self_position = observations[self._self_key]
    target_position = observations[self._target_key]

    # Calculates delta vector and projects it to xy-plane.
    delta_vector = (target_position - self_position) * (1, 1, 0)
    delta_t = time_sec - self._time_sec

    # Advances internal time.
    self._time_sec = time_sec

    if self._verbose:
      with np.printoptions(precision=3, suppress=True):
        logging.info("t = %.1f, self %s: %s, target %s: %s, v: %s, dt %.1f.",
                     self._t,
                     self._self_key, observations[self._self_key],
                     self._target_key, observations[self._target_key],
                     delta_vector, delta_t)

    # Avoids sigularity when it is close enough. Keeps previous orientation.
    distance = np.linalg.norm(delta_vector)
    if distance < _EPS_DISTANCE:
      return target_position.copy(), self._previous_orientation.copy(), {}

    unit_delta_vector = delta_vector / distance
    new_position = (unit_delta_vector * min(self._speed_mps * delta_t, distance)
                    + self_position)

    new_yaw = np.arctan2(unit_delta_vector[1], unit_delta_vector[0])
    new_orientation = np.array((0, 0, np.sin(new_yaw / 2), np.cos(new_yaw / 2)))
    self._previous_orientation = new_orientation

    return new_position, new_orientation.copy(), {}


@gin.configurable
class AnimationFrameController(ControllerBase):
  """An extra action controller to control playback of animation sequence."""

  def __init__(self, fps: float = 10.0,
               pause_between_repeat_sec: float = 0.0,
               start_time_sec: float = 0.0):
    """Constructor.

    Args:
      fps: Frame per second of animation.
      pause_between_repeat_sec: Pause between repeat in second.
      start_time_sec: The time when animation starts to play.
    """
    self._fps = fps
    self._total_length = None
    self._pause_between_repeat_sec = pause_between_repeat_sec
    self._start_time_sec = start_time_sec

  def set_total_length(self, total_length: int):
    """Sets total animation frame length."""
    if total_length <= 0:
      raise ValueError(
          f"Total number of frame must be >= 0, got {total_length}.")
    self._total_length = total_length

  def get_action(self,
                 time_sec: float,
                 observations: Dict[Text, Any]) -> ControllerOutput:
    """Returns animation frame number with default position and orientation.

    Args:
      time_sec: Time since simulation reset in seconds. If time < 0, returns
        initial values and ignores observations.
      observations: A dict of all observations.
    """
    time_sec = max(0, time_sec - self._start_time_sec)
    frame = int(time_sec * self._fps)
    if self._total_length:
      frame = frame % (
          self._total_length + int(self._pause_between_repeat_sec * self._fps))
      frame = min(frame, self._total_length - 1)
    return np.ndarray((0, 0, 0)), np.ndarray((0, 0, 0, 1)), {
        ANIMATION_FRAME_NUMBER_KEY: frame}


@gin.configurable
class ConversationController(ControllerBase):
  """Controller for an object that mimics conversational behavior.

  A controlled object is arrayed in a conversation about a center point.
  When a target object reaches a thresholded distance away from the center,
  the controlled object will face the target object and move away from
  the target's intended path along an orthogonal direction vector until it
  passes.
  """

  def __init__(self,
               self_key: Text,
               target_key: Text,
               position: Sequence[float] = None,
               orientation: Sequence[float] = None,
               conversation_center: Sequence[float] = None,
               proximity_threshold: float = 1.0,
               speed_mps: float = 0.1):
    """Constructor.

    Args:
      self_key: Observation dict key of position of object being controlled.
      target_key: Observation dict key of position of target object.
      position: Initial position of the object.
      orientation: Initial orientation of the object in xyzw quaternion.
      conversation_center: Position of the center of the conversation group.
      proximity_threshold: The distance from the conversation center that the
        target must reach in order to prompt a response from the controlled
        object.
      speed_mps: Speed in meters per second, always positive.
    """
    self._self_key = self_key
    self._target_key = target_key
    self._position = np.array(position or (0, 0, 0))
    self._orientation = np.array(orientation or (0, 0, 0, 1))
    self._conversation_center = np.array(conversation_center or (0, 0, 0))
    self._proximity_threshold = proximity_threshold
    self._speed_mps = speed_mps

    self._prev_target_position = None
    self._wait_position = None

  def _get_wait_position(self, self_position, target_position):
    """Gets the waiting position for the controlled object.

    This returns the position that the controlled object should move towards
    to create physical space such that the target object may pass through
    the conversation space.

    Args:
      self_position: The current position of the controlled object.
      target_position: The current position of the target object.

    Returns:
      An xyz position representing the waiting position that the controlled
      object should move towards to create space.
    """
    # Find an orthogonal projection to the target's path
    target_path = np.array(target_position - self._prev_target_position)
    self_path = np.array(self_position - self._prev_target_position)

    unit_target_path = target_path / np.linalg.norm(target_path)

    projected_point = np.dot(self_path, unit_target_path) * unit_target_path
    projected_point += self._prev_target_position

    # Get the position that lies along the orthogonal projection vector
    # but in the opposite direction from the target's path and exactly
    # proximity_threshold distance away.
    return self._get_position(
        projected_point,
        self_position,
        self._proximity_threshold,
        self._proximity_threshold)

  def _get_orientation(self, source_position, target_position):
    """Gets orientation required to face target_position from source_position.

    Args:
      source_position: The source position where an object would be located.
      target_position: The target position that an object should face.

    Returns:
      A xyzw quaternion indicating the orientation.
    """
    if np.allclose(source_position, target_position):
      return self._orientation

    delta_vector = (target_position - source_position) * (1, 1, 0)

    new_yaw = np.arctan2(delta_vector[1], delta_vector[0])
    new_orientation = np.array(
        (0, 0, np.sin(new_yaw / 2), np.cos(new_yaw / 2)))

    return new_orientation

  def _get_position(self,
                    source_position,
                    target_position,
                    min_delta,
                    max_delta):
    """Gets the next position along the vector from source to target.

    This returns the position that should be moved to next which lies along
    the direction vector from source -> target with a minimum length of
    min_delta and a maximum distance of max_delta.

    Args:
      source_position: The current position of the controlled object.
      target_position: The target position to move to.
      min_delta: The minimum amount of distance to move.
      max_delta: The maximum amount of distance to move.

    Returns:
      An xyz position representing the next position to move to.
    """
    delta_vector = (target_position - source_position) * (1, 1, 0)
    distance = np.linalg.norm(delta_vector)

    # If the distance to the target is greater than the maximum step delta,
    # then normalize the vector and set it to the max step delta.
    if distance > max_delta:
      delta_vector = (delta_vector / distance) * max_delta
    # If the distance is less than the minimum step delta, then normalize
    # the vector and set it to the min step delta.
    elif distance < min_delta:
      delta_vector = (delta_vector / distance) * min_delta

    new_position = (delta_vector + source_position)

    return new_position

  def _get_target_distance_to_center(self, target_position):
    """Gets the distance from the target to the conversation center point.

    Args:
      target_position: The target position.

    Returns:
      The scalar distance from the target position to the conversation center.
    """
    # Calculates delta vector and projects it to xy-plane.
    delta_vector = (target_position - self._conversation_center) * (1, 1, 0)

    # Compute the length of the delta vector.
    return np.linalg.norm(delta_vector)

  def get_action(self,
                 t: float,
                 observations: Dict[Text, Any]) -> ControllerOutput:
    """Gets the position and orientation of the controlled object.

    Args:
      t: The current time step.
      observations: Dict containing sensor observations for current time step.

    Returns:
      The new position and orientation for the controlled object.
    """

    position = self._position
    orientation = self._orientation

    # Observations are only available for positive time steps.
    if t >= 0:
      self_position = observations[self._self_key]
      target_position = observations[self._target_key]

      target_distance = self._get_target_distance_to_center(target_position)

      # Check if the target is within the threshold distance of the
      # conversation center.
      if(target_distance < self._proximity_threshold and
         self._prev_target_position is not None):

        # If it is, get the position that the controlled object should move to
        # in order to create space and get the resulting action/orientation.
        wait_position = self._get_wait_position(self_position, target_position)

        orientation = self._get_orientation(
            self_position,
            target_position)

        position = self._get_position(
            self_position,
            wait_position,
            0.0,
            self._speed_mps)
      else:
        # Otherwise, get the position/orientation required to move back to
        # the original position and face the conversation center.
        orientation = self._get_orientation(
            self_position,
            self._conversation_center)

        position = self._get_position(
            self_position,
            self._position,
            0.0,
            self._speed_mps)

      self._prev_target_position = target_position

    # Only return a new position along the x/y axis, z should be unaffected.
    position = np.array([position[0], position[1], self._position[2]])
    return position, orientation, {}


@gin.configurable
class PauseIfCloseByWrapper(ControllerBase):
  """A controller wrapper that pauses controller if object is close to others.

  This wrapper works best if the underlying controller is time based. It is
  intended to be a simple way to stop agent when blocked and is not for
  reliable collision avoidance.
  """

  _DEFAULT_PAUSE_DISTANCE_M = 1.0

  def __init__(
      self,
      wrapped_controller: ControllerBase,
      self_pos_key: Text,
      others_pos_keys: Sequence[Text],
      pause_distance: Union[float, Sequence[float]] = _DEFAULT_PAUSE_DISTANCE_M,
      self_yaw_key: Optional[Text] = None,
      active_front_sector: Optional[float] = None):
    """Constructor.

    Args:
      wrapped_controller: The controller being wrapped.
      self_pos_key: Observation key of self position.
      others_pos_keys: Observation keys of others' positions.
      pause_distance: The distance limit before the controller pauses in meters.
        Can be a float value which applies to all objects specified in
        others_pos_keys or a Sequence of float values with the same length
        as others_pos_keys denoting the pause distance for each individual
        object in the same order in others_pos_keys. Default pause distance is
        one meter.
      self_yaw_key: Observation key of self yaw. Required if active_front_sector
        is specified.
      active_front_sector: If specified, it defines pie-shaped active region in
        front of controlled object. The pie-shaped area is symmetric about the
        forward direction of controlled object with it angle defined by this
        arg in radians, Only when other objects shows up in this region and
        pause distance requirement is met, pause is actived.
    """

    self._controller = wrapped_controller
    self._pause_start_t = -1
    self._shift_t = 0
    self._last_action = None
    self._self_pos_key = self_pos_key
    self._others_pos_keys = list(others_pos_keys)  # Make a copy.

    if isinstance(pause_distance, float):
      pause_distance = [pause_distance] * len(others_pos_keys)

    if len(pause_distance) != len(others_pos_keys):
      raise ValueError(
          "pause_distance and others_pos_keys must have the same length.")
    self._pause_distance = list(pause_distance)  # Make a copy.

    if active_front_sector is not None and self_yaw_key is None:
      raise ValueError(
          "self_yaw_key must be specified if active_front_sector is specified.")

    self._self_yaw_key = self_yaw_key
    self._active_front_sector = active_front_sector

  def get_action(self,
                 t: float,
                 observations: Dict[Text, Any]) -> ControllerOutput:

    """Gets the position and orientation of the controlled object.

    Args:
      t: The current time step.
      observations: Dict containing sensor observations for current time step.

    Returns:
      The new position and orientation for the controlled object.
    """
    if t < 0:
      self._pause_start_t = -1
      self._shift_t = 0
      self._last_action = self._controller.get_action(
          t, observations)
      return self._last_action

    if self._should_pause(observations):
      # Only record the start time of pause.
      if self._pause_start_t < 0:
        self._pause_start_t = t
      return self._last_action
    else:
      if self._pause_start_t >= 0:
        self._shift_t += t - self._pause_start_t
        self._pause_start_t = -1

      self._last_action = self._controller.get_action(
          t - self._shift_t, observations)
      return self._last_action

  def _should_pause(self, observations) -> bool:
    """Determines whether the controller should pause."""
    self_position_2d = observations[self._self_pos_key][:2]
    for pos_key, pause_distance in zip(
        self._others_pos_keys, self._pause_distance):
      position_2d = observations[pos_key][:2]
      vector_2d = position_2d - self_position_2d
      distance = np.linalg.norm(vector_2d)

      if self._active_front_sector is None:
        return distance <= pause_distance
      else:
        yaw = observations[self._self_yaw_key][0]
        dot = np.dot(vector_2d / distance, np.array([np.cos(yaw), np.sin(yaw)]))
        return (distance <= pause_distance and
                np.arccos(dot) < self._active_front_sector / 2)

    return False



"""Asymmetric sine controller for quadruped locomotion.

Asymmetric sine uses cosine and sine waves to generate swinging and extension
for leg motion. It's asymmetric because sine waves are split into two phases
(swing forward and stance) and these phases have different frequencies according
to what proportion of a period will be spend on swinging forward forward
swinging backwards. In addition, the sine wave for extension has different
amplitudes during these two phases.
"""
import math

TWO_PI = 2 * math.pi
_DEFAULT_LEG_AMPLITUDE_EXTENSION = -0.02
_DEFAULT_LEG_AMPLITUDE_SWING = 0.5
_DEFAULT_LEG_AMPLITUDE_LIFT = 0.9
_DEFAULT_WALKING_HEIGHT = 0.0
_DEFAULT_LEG_CENTER_SWING = 0.0
_DELTA_CENTER_EXTENSION_CAP = 1
_DELTA_INTENSITY_CAP = 0.1


class SimpleLegController(object):
  """Controller that gives swing and extension based on phase and parameters.


  The controller returns the swing-extend pair based on a parameterized
  ellipsoid trajectory that depends on center of motion, amplitude and phase.
  The parameters are
    amplitude_extension: Amplitude for extension during stance (phase < pi).
    amplitude_lift: Amplitude for extension during swing (phase > pi).
    amplitude_swing: Amplitude for swing.
    center_extension: The value extension signal oscillates around.
    center_swing: The value swing signal oscillates around.
    intensity: A coefficient that scales the motion of the legs.
  The formula to calculate motion and more detailed information about these
  parameters can be found at go/pmtg-refactored.
  """

  def __init__(self, init_phase=0):
    self.amplitude_extension = _DEFAULT_LEG_AMPLITUDE_EXTENSION
    self.amplitude_swing = _DEFAULT_LEG_AMPLITUDE_SWING
    self.amplitude_lift = _DEFAULT_LEG_AMPLITUDE_LIFT
    self.center_extension = _DEFAULT_WALKING_HEIGHT
    self.center_swing = _DEFAULT_LEG_CENTER_SWING
    self.intensity = 1.0
    self._init_phase = init_phase
    self.phase = init_phase
    self.phase_offset = 0

  def reset(self):
    self.phase = self._init_phase

  def get_swing_extend(self):
    """Returns the swing and extend parameters for the leg.

    Returns:
      swing: Desired swing of the leg.
      extend: Desired extension amount of the leg.
    """

    # Increase default extension by the extra extension scaled by intensity.
    # Extend reduces to default center extension when intensity goes to 0,
    # because we prefer the legs to stay at walking height when intensity is
    # 0.
    amplitude_extension = self.amplitude_extension
    # The leg is in swing phase when phase > pi.
    if self.phase > math.pi:
      amplitude_extension = self.amplitude_lift
    extend = self.center_extension + (
        amplitude_extension * math.sin(self.phase)) * self.intensity
    # Calculate the swing based on the signal and scale it with intensity.
    # Swing reduces to 0 when intensity goes to 0, because we would prefer the
    # legs to stay neutral (standing position instead of center swing) when
    # intensity is 0.
    swing = self.center_swing + self.amplitude_swing * math.cos(self.phase)
    swing *= self.intensity

    return swing, extend

  def adjust_center_extension(self, target_center_extension):
    delta = min(_DELTA_CENTER_EXTENSION_CAP,
                target_center_extension - self.center_extension)
    self.center_extension += delta

  def adjust_intensity(self, target_intensity):
    delta = min(_DELTA_INTENSITY_CAP, target_intensity - self.intensity)
    self.intensity += delta



# Lint as: python3
"""Scene with randomly spaced stepstones."""

from typing import Optional, Sequence

import gin
import numpy as np

from pybullet_envs.minitaur.envs_v2.scenes import scene_base
from pybullet_envs.minitaur.envs_v2.scenes.terrain import stepstones


@gin.configurable
class RandomStepstoneScene(scene_base.SceneBase):
  """Scene with randomly spaced stepstones."""

  def __init__(
      self,
      num_stones: int = 50,
      stone_height: float = 0.1,
      stone_width_lower_bound: float = 10.0,
      stone_width_upper_bound: float = 10.0,
      stone_length_lower_bound: float = 0.1,
      stone_length_upper_bound: float = 0.3,
      gap_length_lower_bound: float = 0.1,
      gap_length_upper_bound: float = 0.3,
      height_offset_lower_bound: float = 0.0,
      height_offset_upper_bound: float = 1e-6,
      floor_height_lower_bound: float = -0.5,
      floor_height_upper_bound: float = -0.55,
      platform_length_lower_bound: float = 0.75,
      platform_length_upper_bound: float = 1.0,
      random_seed: Optional[int] = None,
      color_sequence: Sequence[Sequence[float]] = stepstones.MULTICOLOR,
      rebuild_scene_during_reset: bool = True):
    """Initializes RandomStepstoneScene.

    Args:
      num_stones: The number of stepstones.
      stone_height: The height in meters of each stepstone.
      stone_width_lower_bound: The lower bound in meters of the randomly sampled
        stepstone width.
      stone_width_upper_bound: The upper bound in meters of the randomly sampled
        stepstone width.
      stone_length_lower_bound: The lower bound in meters of the randomly
        sampled stepstone length.
      stone_length_upper_bound: The upper bound in meters of the randomly
        sampled stepstone length.
      gap_length_lower_bound: The lower bound in meters of the random sampled
        gap distance.
      gap_length_upper_bound: The upper bound in meters of the random sampled
        gap distance.
      height_offset_lower_bound: The lower bound in meters of the randomly
        sampled stepstone height.
      height_offset_upper_bound: The upper bound in meters of the randomly
        sampled stepstone height.
      floor_height_lower_bound: The lower bound in meters of the randomly
        sampled floor height.
      floor_height_upper_bound: The upper bound in meters of the randomly
        sampled floor height.
      platform_length_lower_bound: The lower bound in meters of the first step
        stone length.
      platform_length_upper_bound: The upper bound in meters of the first step
        stone length.
      random_seed: The random seed to generate the random stepstones.
      color_sequence: A list of (red, green, blue, alpha) colors where each
        element is in [0, 1] and alpha is transparency. The stepstones will
        cycle through these colors.
      rebuild_scene_during_reset: Whether to rebuild the stepstones during
        reset.
    """
    for color in color_sequence:
      if len(color) != 4:
        raise ValueError(
            "Each color must be length 4; got <{}>".format(color_sequence))

    super(RandomStepstoneScene, self).__init__(data_root=None)
    self._num_stones = num_stones
    self._stone_height = stone_height
    self._stone_width_lower_bound = stone_width_lower_bound
    self._stone_width_upper_bound = stone_width_upper_bound
    self._stone_length_lower_bound = stone_length_lower_bound
    self._stone_length_upper_bound = stone_length_upper_bound
    self._gap_length_lower_bound = gap_length_lower_bound
    self._gap_length_upper_bound = gap_length_upper_bound
    self._height_offset_lower_bound = height_offset_lower_bound
    self._height_offset_upper_bound = height_offset_upper_bound
    self._floor_height_lower_bound = floor_height_lower_bound
    self._floor_height_upper_bound = floor_height_upper_bound
    self._platform_length_lower_bound = platform_length_lower_bound
    self._platform_length_upper_bound = platform_length_upper_bound
    self._random_seed = random_seed
    self._color_sequence = color_sequence
    self._rebuild_scene_during_reset = rebuild_scene_during_reset

  def reset(self):
    super().reset()

    if self._rebuild_scene_during_reset:
      for ground_id in self.ground_ids:
        self._pybullet_client.removeBody(ground_id)
      self.build_scene(self._pybullet_client)

  def build_scene(self, pybullet_client):
    super().build_scene(pybullet_client)
    # The first stone is to let the robot stand at the initial position.

    stone_width = np.random.uniform(self._stone_width_lower_bound,
                                    self._stone_width_upper_bound)
    platform_length = np.random.uniform(self._platform_length_lower_bound,
                                        self._platform_length_upper_bound)

    end_pos, first_stone_id = stepstones.build_one_stepstone(
        pybullet_client=pybullet_client,
        start_pos=(-platform_length / 2.0, 0, 0),
        stone_length=platform_length,
        stone_height=self._stone_height,
        stone_width=stone_width,
        gap_length=0.0,
        height_offset=0.0,
        rgba_color=stepstones.GRAY)

    _, stepstone_ids = stepstones.build_random_stepstones(
        pybullet_client=self.pybullet_client,
        start_pos=end_pos,
        num_stones=self._num_stones,
        stone_height=self._stone_height,
        stone_width=stone_width,
        stone_length_lower_bound=self._stone_length_lower_bound,
        stone_length_upper_bound=self._stone_length_upper_bound,
        gap_length_lower_bound=self._gap_length_lower_bound,
        gap_length_upper_bound=self._gap_length_upper_bound,
        height_offset_lower_bound=self._height_offset_lower_bound,
        height_offset_upper_bound=self._height_offset_upper_bound,
        random_seed=self._random_seed,
        color_sequence=self._color_sequence)
    for pybullet_id in [first_stone_id] + stepstone_ids:
      self.add_object(pybullet_id, scene_base.ObjectType.GROUND)

    self._floor_height = np.random.uniform(self._floor_height_lower_bound,
                                           self._floor_height_upper_bound)
    # Add floor
    floor_id = stepstones.load_box(pybullet_client,
        half_extents=[100, 100, 1],
        position=np.array([0.0, 0.0, self._floor_height - 1.0]),
        orientation=(0.0, 0.0, 0.0, 1.0),
        rgba_color=(1.0, 1.0, 1.0, 1.0),
        mass=0)
    self.add_object(floor_id, scene_base.ObjectType.GROUND)



import pybullet as p
import time
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
cartpole = p.loadURDF("cartpole.urdf")
p.setRealTimeSimulation(1)
p.setJointMotorControl2(cartpole,
                        1,
                        p.POSITION_CONTROL,
                        targetPosition=1000,
                        targetVelocity=0,
                        force=1000,
                        positionGain=1,
                        velocityGain=0,
                        maxVelocity=0.5)
while (1):
  p.setGravity(0, 0, -10)
  js = p.getJointState(cartpole, 1)
  print("position=", js[0], "velocity=", js[1])
  time.sleep(0.01)



from pybullet_utils import bullet_client
import math


class HumanoidPoseInterpolator(object):

  def __init__(self):
    pass

  def Reset(self,
            basePos=[0, 0, 0],
            baseOrn=[0, 0, 0, 1],
            chestRot=[0, 0, 0, 1],
            neckRot=[0, 0, 0, 1],
            rightHipRot=[0, 0, 0, 1],
            rightKneeRot=[0],
            rightAnkleRot=[0, 0, 0, 1],
            rightShoulderRot=[0, 0, 0, 1],
            rightElbowRot=[0],
            leftHipRot=[0, 0, 0, 1],
            leftKneeRot=[0],
            leftAnkleRot=[0, 0, 0, 1],
            leftShoulderRot=[0, 0, 0, 1],
            leftElbowRot=[0],
            baseLinVel=[0, 0, 0],
            baseAngVel=[0, 0, 0],
            chestVel=[0, 0, 0],
            neckVel=[0, 0, 0],
            rightHipVel=[0, 0, 0],
            rightKneeVel=[0],
            rightAnkleVel=[0, 0, 0],
            rightShoulderVel=[0, 0, 0],
            rightElbowVel=[0],
            leftHipVel=[0, 0, 0],
            leftKneeVel=[0],
            leftAnkleVel=[0, 0, 0],
            leftShoulderVel=[0, 0, 0],
            leftElbowVel=[0]):

    self._basePos = basePos
    self._baseLinVel = baseLinVel
    #print("HumanoidPoseInterpolator.Reset: baseLinVel = ", baseLinVel)
    self._baseOrn = baseOrn
    self._baseAngVel = baseAngVel

    self._chestRot = chestRot
    self._chestVel = chestVel
    self._neckRot = neckRot
    self._neckVel = neckVel

    self._rightHipRot = rightHipRot
    self._rightHipVel = rightHipVel
    self._rightKneeRot = rightKneeRot
    self._rightKneeVel = rightKneeVel
    self._rightAnkleRot = rightAnkleRot
    self._rightAnkleVel = rightAnkleVel

    self._rightShoulderRot = rightShoulderRot
    self._rightShoulderVel = rightShoulderVel
    self._rightElbowRot = rightElbowRot
    self._rightElbowVel = rightElbowVel

    self._leftHipRot = leftHipRot
    self._leftHipVel = leftHipVel
    self._leftKneeRot = leftKneeRot
    self._leftKneeVel = leftKneeVel
    self._leftAnkleRot = leftAnkleRot
    self._leftAnkleVel = leftAnkleVel

    self._leftShoulderRot = leftShoulderRot
    self._leftShoulderVel = leftShoulderVel
    self._leftElbowRot = leftElbowRot
    self._leftElbowVel = leftElbowVel

  def ComputeLinVel(self, posStart, posEnd, deltaTime):
    vel = [(posEnd[0] - posStart[0]) / deltaTime, (posEnd[1] - posStart[1]) / deltaTime,
           (posEnd[2] - posStart[2]) / deltaTime]
    return vel

  def ComputeAngVel(self, ornStart, ornEnd, deltaTime, bullet_client):
    dorn = bullet_client.getDifferenceQuaternion(ornStart, ornEnd)
    axis, angle = bullet_client.getAxisAngleFromQuaternion(dorn)
    angVel = [(axis[0] * angle) / deltaTime, (axis[1] * angle) / deltaTime,
              (axis[2] * angle) / deltaTime]
    return angVel

  def ComputeAngVelRel(self, ornStart, ornEnd, deltaTime, bullet_client):
    ornStartConjugate = [-ornStart[0], -ornStart[1], -ornStart[2], ornStart[3]]
    pos_diff, q_diff = bullet_client.multiplyTransforms([0, 0, 0], ornStartConjugate, [0, 0, 0],
                                                        ornEnd)
    axis, angle = bullet_client.getAxisAngleFromQuaternion(q_diff)
    angVel = [(axis[0] * angle) / deltaTime, (axis[1] * angle) / deltaTime,
              (axis[2] * angle) / deltaTime]
    return angVel

  def NormalizeVector(self, vec):
    length2 = orn[0] * orn[0] + orn[1] * orn[1] + orn[2] * orn[2]
    if (length2 > 0):
      length = math.sqrt(length2)

  def NormalizeQuaternion(self, orn):
    length2 = orn[0] * orn[0] + orn[1] * orn[1] + orn[2] * orn[2] + orn[3] * orn[3]
    if (length2 > 0):
      length = math.sqrt(length2)
      orn[0] /= length
      orn[1] /= length
      orn[2] /= length
      orn[3] /= length
      return orn

    #print("Normalize? length=",length)

  def PostProcessMotionData(self, frameData):
    baseOrn1Start = [frameData[5], frameData[6], frameData[7], frameData[4]]

    chestRotStart = [frameData[9], frameData[10], frameData[11], frameData[8]]

    neckRotStart = [frameData[13], frameData[14], frameData[15], frameData[12]]
    rightHipRotStart = [frameData[17], frameData[18], frameData[19], frameData[16]]
    rightAnkleRotStart = [frameData[22], frameData[23], frameData[24], frameData[21]]
    rightShoulderRotStart = [frameData[26], frameData[27], frameData[28], frameData[25]]
    leftHipRotStart = [frameData[31], frameData[32], frameData[33], frameData[30]]
    leftAnkleRotStart = [frameData[36], frameData[37], frameData[38], frameData[35]]
    leftShoulderRotStart = [frameData[40], frameData[41], frameData[42], frameData[39]]

  def GetPose(self):
    pose = [
        self._basePos[0], self._basePos[1], self._basePos[2], self._baseOrn[0], self._baseOrn[1],
        self._baseOrn[2], self._baseOrn[3], self._chestRot[0], self._chestRot[1],
        self._chestRot[2], self._chestRot[3], self._neckRot[0], self._neckRot[1], self._neckRot[2],
        self._neckRot[3], self._rightHipRot[0], self._rightHipRot[1], self._rightHipRot[2],
        self._rightHipRot[3], self._rightKneeRot[0], self._rightAnkleRot[0],
        self._rightAnkleRot[1], self._rightAnkleRot[2], self._rightAnkleRot[3],
        self._rightShoulderRot[0], self._rightShoulderRot[1], self._rightShoulderRot[2],
        self._rightShoulderRot[3], self._rightElbowRot[0], self._leftHipRot[0],
        self._leftHipRot[1], self._leftHipRot[2], self._leftHipRot[3], self._leftKneeRot[0],
        self._leftAnkleRot[0], self._leftAnkleRot[1], self._leftAnkleRot[2], self._leftAnkleRot[3],
        self._leftShoulderRot[0], self._leftShoulderRot[1], self._leftShoulderRot[2],
        self._leftShoulderRot[3], self._leftElbowRot[0]
    ]
    return pose

  def Slerp(self, frameFraction, frameData, frameDataNext, bullet_client):
    keyFrameDuration = frameData[0]
    basePos1Start = [frameData[1], frameData[2], frameData[3]]
    basePos1End = [frameDataNext[1], frameDataNext[2], frameDataNext[3]]
    self._basePos = [
        basePos1Start[0] + frameFraction * (basePos1End[0] - basePos1Start[0]),
        basePos1Start[1] + frameFraction * (basePos1End[1] - basePos1Start[1]),
        basePos1Start[2] + frameFraction * (basePos1End[2] - basePos1Start[2])
    ]
    self._baseLinVel = self.ComputeLinVel(basePos1Start, basePos1End, keyFrameDuration)
    baseOrn1Start = [frameData[5], frameData[6], frameData[7], frameData[4]]
    baseOrn1Next = [frameDataNext[5], frameDataNext[6], frameDataNext[7], frameDataNext[4]]
    self._baseOrn = bullet_client.getQuaternionSlerp(baseOrn1Start, baseOrn1Next, frameFraction)
    self._baseAngVel = self.ComputeAngVel(baseOrn1Start, baseOrn1Next, keyFrameDuration,
                                          bullet_client)

    ##pre-rotate to make z-up
    #y2zPos=[0,0,0.0]
    #y2zOrn = p.getQuaternionFromEuler([1.57,0,0])
    #basePos,baseOrn = p.multiplyTransforms(y2zPos, y2zOrn,basePos1,baseOrn1)

    chestRotStart = [frameData[9], frameData[10], frameData[11], frameData[8]]
    chestRotEnd = [frameDataNext[9], frameDataNext[10], frameDataNext[11], frameDataNext[8]]
    self._chestRot = bullet_client.getQuaternionSlerp(chestRotStart, chestRotEnd, frameFraction)
    self._chestVel = self.ComputeAngVelRel(chestRotStart, chestRotEnd, keyFrameDuration,
                                           bullet_client)

    neckRotStart = [frameData[13], frameData[14], frameData[15], frameData[12]]
    neckRotEnd = [frameDataNext[13], frameDataNext[14], frameDataNext[15], frameDataNext[12]]
    self._neckRot = bullet_client.getQuaternionSlerp(neckRotStart, neckRotEnd, frameFraction)
    self._neckVel = self.ComputeAngVelRel(neckRotStart, neckRotEnd, keyFrameDuration,
                                          bullet_client)

    rightHipRotStart = [frameData[17], frameData[18], frameData[19], frameData[16]]
    rightHipRotEnd = [frameDataNext[17], frameDataNext[18], frameDataNext[19], frameDataNext[16]]
    self._rightHipRot = bullet_client.getQuaternionSlerp(rightHipRotStart, rightHipRotEnd,
                                                         frameFraction)
    self._rightHipVel = self.ComputeAngVelRel(rightHipRotStart, rightHipRotEnd, keyFrameDuration,
                                              bullet_client)

    rightKneeRotStart = [frameData[20]]
    rightKneeRotEnd = [frameDataNext[20]]
    self._rightKneeRot = [
        rightKneeRotStart[0] + frameFraction * (rightKneeRotEnd[0] - rightKneeRotStart[0])
    ]
    self._rightKneeVel = [(rightKneeRotEnd[0] - rightKneeRotStart[0]) / keyFrameDuration]

    rightAnkleRotStart = [frameData[22], frameData[23], frameData[24], frameData[21]]
    rightAnkleRotEnd = [frameDataNext[22], frameDataNext[23], frameDataNext[24], frameDataNext[21]]
    self._rightAnkleRot = bullet_client.getQuaternionSlerp(rightAnkleRotStart, rightAnkleRotEnd,
                                                           frameFraction)
    self._rightAnkleVel = self.ComputeAngVelRel(rightAnkleRotStart, rightAnkleRotEnd,
                                                keyFrameDuration, bullet_client)

    rightShoulderRotStart = [frameData[26], frameData[27], frameData[28], frameData[25]]
    rightShoulderRotEnd = [
        frameDataNext[26], frameDataNext[27], frameDataNext[28], frameDataNext[25]
    ]
    self._rightShoulderRot = bullet_client.getQuaternionSlerp(rightShoulderRotStart,
                                                              rightShoulderRotEnd, frameFraction)
    self._rightShoulderVel = self.ComputeAngVelRel(rightShoulderRotStart, rightShoulderRotEnd,
                                                   keyFrameDuration, bullet_client)

    rightElbowRotStart = [frameData[29]]
    rightElbowRotEnd = [frameDataNext[29]]
    self._rightElbowRot = [
        rightElbowRotStart[0] + frameFraction * (rightElbowRotEnd[0] - rightElbowRotStart[0])
    ]
    self._rightElbowVel = [(rightElbowRotEnd[0] - rightElbowRotStart[0]) / keyFrameDuration]

    leftHipRotStart = [frameData[31], frameData[32], frameData[33], frameData[30]]
    leftHipRotEnd = [frameDataNext[31], frameDataNext[32], frameDataNext[33], frameDataNext[30]]
    self._leftHipRot = bullet_client.getQuaternionSlerp(leftHipRotStart, leftHipRotEnd,
                                                        frameFraction)
    self._leftHipVel = self.ComputeAngVelRel(leftHipRotStart, leftHipRotEnd, keyFrameDuration,
                                             bullet_client)

    leftKneeRotStart = [frameData[34]]
    leftKneeRotEnd = [frameDataNext[34]]
    self._leftKneeRot = [
        leftKneeRotStart[0] + frameFraction * (leftKneeRotEnd[0] - leftKneeRotStart[0])
    ]
    self._leftKneeVel = [(leftKneeRotEnd[0] - leftKneeRotStart[0]) / keyFrameDuration]

    leftAnkleRotStart = [frameData[36], frameData[37], frameData[38], frameData[35]]
    leftAnkleRotEnd = [frameDataNext[36], frameDataNext[37], frameDataNext[38], frameDataNext[35]]
    self._leftAnkleRot = bullet_client.getQuaternionSlerp(leftAnkleRotStart, leftAnkleRotEnd,
                                                          frameFraction)
    self._leftAnkleVel = self.ComputeAngVelRel(leftAnkleRotStart, leftAnkleRotEnd,
                                               keyFrameDuration, bullet_client)

    leftShoulderRotStart = [frameData[40], frameData[41], frameData[42], frameData[39]]
    leftShoulderRotEnd = [
        frameDataNext[40], frameDataNext[41], frameDataNext[42], frameDataNext[39]
    ]
    self._leftShoulderRot = bullet_client.getQuaternionSlerp(leftShoulderRotStart,
                                                             leftShoulderRotEnd, frameFraction)
    self._leftShoulderVel = self.ComputeAngVelRel(leftShoulderRotStart, leftShoulderRotEnd,
                                                  keyFrameDuration, bullet_client)

    leftElbowRotStart = [frameData[43]]
    leftElbowRotEnd = [frameDataNext[43]]
    self._leftElbowRot = [
        leftElbowRotStart[0] + frameFraction * (leftElbowRotEnd[0] - leftElbowRotStart[0])
    ]
    self._leftElbowVel = [(leftElbowRotEnd[0] - leftElbowRotStart[0]) / keyFrameDuration]

    pose = self.GetPose()
    return pose

  def ConvertFromAction(self, pybullet_client, action):
    #turn action into pose

    self.Reset()  #?? needed?
    index = 0
    angle = action[index]
    axis = [action[index + 1], action[index + 2], action[index + 3]]
    index += 4
    self._chestRot = pybullet_client.getQuaternionFromAxisAngle(axis, angle)
    #print("pose._chestRot=",pose._chestRot)

    angle = action[index]
    axis = [action[index + 1], action[index + 2], action[index + 3]]
    index += 4
    self._neckRot = pybullet_client.getQuaternionFromAxisAngle(axis, angle)

    angle = action[index]
    axis = [action[index + 1], action[index + 2], action[index + 3]]
    index += 4
    self._rightHipRot = pybullet_client.getQuaternionFromAxisAngle(axis, angle)

    angle = action[index]
    index += 1
    self._rightKneeRot = [angle]

    angle = action[index]
    axis = [action[index + 1], action[index + 2], action[index + 3]]
    index += 4
    self._rightAnkleRot = pybullet_client.getQuaternionFromAxisAngle(axis, angle)

    angle = action[index]
    axis = [action[index + 1], action[index + 2], action[index + 3]]
    index += 4
    self._rightShoulderRot = pybullet_client.getQuaternionFromAxisAngle(axis, angle)

    angle = action[index]
    index += 1
    self._rightElbowRot = [angle]

    angle = action[index]
    axis = [action[index + 1], action[index + 2], action[index + 3]]
    index += 4
    self._leftHipRot = pybullet_client.getQuaternionFromAxisAngle(axis, angle)

    angle = action[index]
    index += 1
    self._leftKneeRot = [angle]

    angle = action[index]
    axis = [action[index + 1], action[index + 2], action[index + 3]]
    index += 4
    self._leftAnkleRot = pybullet_client.getQuaternionFromAxisAngle(axis, angle)

    angle = action[index]
    axis = [action[index + 1], action[index + 2], action[index + 3]]
    index += 4
    self._leftShoulderRot = pybullet_client.getQuaternionFromAxisAngle(axis, angle)

    angle = action[index]
    index += 1
    self._leftElbowRot = [angle]

    pose = self.GetPose()
    return pose



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the weight counting utility."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tf.compat.v1 as tf

from agents.tools import count_weights


class CountWeightsTest(tf.test.TestCase):

  def test_count_trainable(self):
    tf.Variable(tf.zeros((5, 3)), trainable=True)
    tf.Variable(tf.zeros((1, 1)), trainable=True)
    tf.Variable(tf.zeros((5,)), trainable=True)
    self.assertEqual(15 + 1 + 5, count_weights())

  def test_ignore_non_trainable(self):
    tf.Variable(tf.zeros((5, 3)), trainable=False)
    tf.Variable(tf.zeros((1, 1)), trainable=False)
    tf.Variable(tf.zeros((5,)), trainable=False)
    self.assertEqual(0, count_weights())

  def test_trainable_and_non_trainable(self):
    tf.Variable(tf.zeros((5, 3)), trainable=True)
    tf.Variable(tf.zeros((8, 2)), trainable=False)
    tf.Variable(tf.zeros((1, 1)), trainable=True)
    tf.Variable(tf.zeros((5,)), trainable=True)
    tf.Variable(tf.zeros((3, 1)), trainable=False)
    self.assertEqual(15 + 1 + 5, count_weights())

  def test_include_scopes(self):
    tf.Variable(tf.zeros((3, 2)), trainable=True)
    with tf.variable_scope('foo'):
      tf.Variable(tf.zeros((5, 2)), trainable=True)
    self.assertEqual(6 + 10, count_weights())

  def test_restrict_scope(self):
    tf.Variable(tf.zeros((3, 2)), trainable=True)
    with tf.variable_scope('foo'):
      tf.Variable(tf.zeros((5, 2)), trainable=True)
      with tf.variable_scope('bar'):
        tf.Variable(tf.zeros((1, 2)), trainable=True)
    self.assertEqual(10 + 2, count_weights('foo'))

  def test_restrict_nested_scope(self):
    tf.Variable(tf.zeros((3, 2)), trainable=True)
    with tf.variable_scope('foo'):
      tf.Variable(tf.zeros((5, 2)), trainable=True)
      with tf.variable_scope('bar'):
        tf.Variable(tf.zeros((1, 2)), trainable=True)
    self.assertEqual(2, count_weights('foo/bar'))

  def test_restrict_invalid_scope(self):
    tf.Variable(tf.zeros((3, 2)), trainable=True)
    with tf.variable_scope('foo'):
      tf.Variable(tf.zeros((5, 2)), trainable=True)
      with tf.variable_scope('bar'):
        tf.Variable(tf.zeros((1, 2)), trainable=True)
    self.assertEqual(0, count_weights('bar'))

  def test_exclude_by_regex(self):
    tf.Variable(tf.zeros((3, 2)), trainable=True)
    with tf.variable_scope('foo'):
      tf.Variable(tf.zeros((5, 2)), trainable=True)
      with tf.variable_scope('bar'):
        tf.Variable(tf.zeros((1, 2)), trainable=True)
    self.assertEqual(0, count_weights(exclude=r'.*'))
    self.assertEqual(6, count_weights(exclude=r'(^|/)foo/.*'))
    self.assertEqual(16, count_weights(exclude=r'.*/bar/.*'))

  def test_non_default_graph(self):
    graph = tf.Graph()
    with graph.as_default():
      tf.Variable(tf.zeros((5, 3)), trainable=True)
      tf.Variable(tf.zeros((8, 2)), trainable=False)
    self.assertNotEqual(graph, tf.get_default_graph)
    self.assertEqual(15, count_weights(graph=graph))


if __name__ == '__main__':
  tf.test.main()



"""This file implements the gym environment of minitaur.

"""
import math
from operator import add
import time

from gym import spaces
import numpy as np
from pybullet_envs.minitaur.envs import minitaur_gym_env
import pybullet_data

ACTION_EPS = 0.01
# RANGE_OF_LEG_MOTION defines how far legs can rotate in both directions
# (1.0 means rotation pi/2 (radians) in both directions).
RANGE_OF_LEG_MOTION = 0.7
# LIMIT_FALLEN defines when to consider robot fallen to the ground.
# This is the vertical component of the robot's front vector (unit vector).
# 0.0 represents body of the robot being horizontal, 1.0 means vertical.
LIMIT_FALLEN = 0.7


class MinitaurStandGymEnv(minitaur_gym_env.MinitaurGymEnv):
  """The gym environment for the minitaur and a ball.

  It simulates the standing up behavior of a minitaur, a quadruped robot. The
  state space include the angles, velocities and torques for all the motors and
  the action space is the desired motor angle for each motor. The reward
  function is based on how long the minitaur stays standing up.

  """
  metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 50}

  def __init__(self,
               urdf_root=pybullet_data.getDataPath(),
               action_repeat=1,
               observation_noise_stdev=minitaur_gym_env.SENSOR_NOISE_STDDEV,
               self_collision_enabled=True,
               motor_velocity_limit=np.inf,
               pd_control_enabled=False,
               render=False):
    """Initialize the minitaur standing up gym environment.

    Args:
      urdf_root: The path to the urdf data folder.
      action_repeat: The number of simulation steps before actions are applied.
      observation_noise_stdev: The standard deviation of observation noise.
      self_collision_enabled: Whether to enable self collision in the sim.
      motor_velocity_limit: The velocity limit of each motor.
      pd_control_enabled: Whether to use PD controller for each motor.
      render: Whether to render the simulation.
    """
    super(MinitaurStandGymEnv, self).__init__(urdf_root=urdf_root,
                                              action_repeat=action_repeat,
                                              observation_noise_stdev=observation_noise_stdev,
                                              self_collision_enabled=self_collision_enabled,
                                              motor_velocity_limit=motor_velocity_limit,
                                              pd_control_enabled=pd_control_enabled,
                                              accurate_motor_model_enabled=True,
                                              motor_overheat_protection=True,
                                              render=render)
    # Set the action dimension to 1, and reset the action space.
    action_dim = 1
    action_high = np.array([self._action_bound] * action_dim)
    self.action_space = spaces.Box(-action_high, action_high)

  def _stand_up(self):
    """Make the robot stand up to its two legs when started on 4 legs.

    This method is similar to the step function, but instead of using the action
    provided, it uses a hand-coded policy to make the robot stand up to its
    two legs. Once the robot is vertical enough it exits and leaves the
    environment to the typical step function that uses agent's actions.

    Returns:
      observations: The angles, velocities and torques of all motors.
      reward: The reward for the current state-action pair.
      done: Whether the episode has ended.
      info: A dictionary that stores diagnostic information.
    """
    for t in range(5000):
      if self._is_render:
        base_pos = self.minitaur.GetBasePosition()
        self._pybullet_client.resetDebugVisualizerCamera(self._cam_dist, self._cam_yaw,
                                                         self._cam_pitch, base_pos)
      state = self._get_true_observation()
      action = self._policy_flip(t, state[24:28])
      self.minitaur.ApplyAction(action)
      self._pybullet_client.stepSimulation()
      self.minitaur.ReceiveObservation()
      if self._is_render:
        time.sleep(self._time_step)
      self._env_step_counter += 1
      reward = self._reward()
      if reward > 0.9:
        break
    done = self._termination()
    return np.array(self._get_observation()), reward, done, {}

  def step(self, action):
    # At start, use policy_flip to lift the robot to its two legs. After the
    # robot reaches the lift up stage, give control back to the agent by
    # returning the current state and reward.
    if self._env_step_counter < 1:
      return self._stand_up()
    return super(MinitaurStandGymEnv, self).step(action)

  def _reward(self):
    """Reward function for standing up pose.

    Returns:
      reward: A number between -1 and 1 according to how vertical is the body of
        the robot.
    """
    orientation = self.minitaur.GetBaseOrientation()
    rot_matrix = self._pybullet_client.getMatrixFromQuaternion(orientation)
    local_front_vec = rot_matrix[6:9]
    alignment = abs(np.dot(np.asarray([1, 0, 0]), np.asarray(local_front_vec)))
    return alignment**4

  def _termination(self):
    if self._is_horizontal():
      return True
    return False

  def _is_horizontal(self):
    """Decide whether minitaur is almost parallel to the ground.

    This method is used in experiments where the robot is learning to stand up.

    Returns:
      Boolean value that indicates whether the minitaur is almost parallel to
      the ground.
    """
    orientation = self.minitaur.GetBaseOrientation()
    rot_matrix = self._pybullet_client.getMatrixFromQuaternion(orientation)
    front_z_component = rot_matrix[6]
    return abs(front_z_component) < LIMIT_FALLEN

  def _transform_action_to_motor_command(self, action):
    """Method to transform the one dimensional action to rotate bottom two legs.

    Args:
      action: A double between -1 and 1, where 0 means keep the legs parallel
        to the body.
    Returns:
      actions: The angles for all motors.
    Raises:
      ValueError: The action dimension is not the same as the number of motors.
      ValueError: The magnitude of actions is out of bounds.
    """
    action = action[0]
    # Scale the action from [-1 to 1] to [-range to +range] (angle in radians).
    action *= RANGE_OF_LEG_MOTION
    action_all_legs = [
        math.pi,  # Upper leg pointing up.
        0,
        0,  # Bottom leg pointing down.
        math.pi,
        0,  # Upper leg pointing up.
        math.pi,
        math.pi,  # Bottom leg pointing down.
        0
    ]
    action_all_legs = [angle - 0.7 for angle in action_all_legs]

    # Use the one dimensional action to rotate both bottom legs.
    action_delta = [0, 0, -action, action, 0, 0, action, -action]
    action_all_legs = map(add, action_all_legs, action_delta)
    return list(action_all_legs)

  def _policy_flip(self, time_step, orientation):
    """Hand coded policy to make the minitaur stand up to its two legs.

    This method is the hand coded policy that uses sine waves and orientation
    of the robot to make it stand up to its two legs. It is composed of these
    behaviors:
    - Rotate bottom legs to always point to the ground
    - Rotate upper legs the other direction so that they point to the sky when
    the robot is standing up, and they point to the ground when the robot is
    horizontal.
    - Shorten the bottom 2 legs
    - Shorten the other two legs, then when the sine wave hits its maximum,
    extend the legs pushing the robot up.

    Args:
      time_step: The time (or frame number) used for sine function.
      orientation: Quaternion specifying the orientation of the body.

    Returns:
      actions: The angles for all motors.
    """
    # Set the default behavior (stand on 4 short legs).
    shorten = -0.7
    a0 = math.pi / 2 + shorten
    a1 = math.pi / 2 + shorten
    a2 = math.pi / 2 + shorten
    a3 = math.pi / 2 + shorten
    a4 = math.pi / 2 + shorten
    a5 = math.pi / 2 + shorten
    a6 = math.pi / 2 + shorten
    a7 = math.pi / 2 + shorten

    # Rotate back legs to point to the ground independent of the orientation.
    rot_matrix = self._pybullet_client.getMatrixFromQuaternion(orientation)
    local_up = rot_matrix[6:]
    multiplier = np.dot(np.asarray([0, 0, 1]), np.asarray(local_up))
    a2 -= (1 - multiplier) * (math.pi / 2)
    a3 += (1 - multiplier) * (math.pi / 2)
    a6 += (1 - multiplier) * (math.pi / 2)
    a7 -= (1 - multiplier) * (math.pi / 2)

    # Rotate front legs the other direction, so that it points up when standing.
    a0 += (1 - multiplier) * (math.pi / 2)
    a1 -= (1 - multiplier) * (math.pi / 2)
    a4 -= (1 - multiplier) * (math.pi / 2)
    a5 += (1 - multiplier) * (math.pi / 2)

    # Periodically push the upper legs to stand up.
    speed = 0.01
    intensity = 1.9
    # Lower the signal a little, so that it becomes positive only for a short
    # amount time.
    lower_signal = -0.94
    signal_unit = math.copysign(intensity, math.sin(time_step * speed) + lower_signal)
    # Only extend the leg, don't shorten.
    if signal_unit < 0:
      signal_unit = 0

    # Only apply it after some time.
    if time_step * speed > math.pi:
      a0 += signal_unit
      a1 += signal_unit
      a4 += signal_unit
      a5 += signal_unit
    joint_values = [a0, a1, a2, a3, a4, a5, a6, a7]
    return joint_values



from abc import abstractmethod
import sys, abc
if sys.version_info >= (3, 4):
  ABC = abc.ABC
else:
  ABC = abc.ABCMeta('ABC', (), {})

import numpy as np
from enum import Enum


class Env(ABC):

  class Terminate(Enum):
    Null = 0
    Fail = 1
    Succ = 2

  def __init__(self, args, enable_draw):
    self.enable_draw = enable_draw
    return



import pybullet as bullet
import pybullet_data as pd

plot = True
import time

if (plot):
  import matplotlib.pyplot as plt
import math
verbose = False

# Parameters:
robot_base = [0., 0., 0.]
robot_orientation = [0., 0., 0., 1.]
delta_t = 0.0001

# Initialize Bullet Simulator
id_simulator = bullet.connect(bullet.GUI)  # or bullet.DIRECT for non-graphical version
bullet.setTimeStep(delta_t)
bullet.setAdditionalSearchPath(pd.getDataPath())

# Switch between URDF with/without FIXED joints
with_fixed_joints = True

if with_fixed_joints:
  id_revolute_joints = [0, 3]
  id_robot = bullet.loadURDF("TwoJointRobot_w_fixedJoints.urdf",
                             robot_base,
                             robot_orientation,
                             useFixedBase=True)
else:
  id_revolute_joints = [0, 1]
  id_robot = bullet.loadURDF("TwoJointRobot_wo_fixedJoints.urdf",
                             robot_base,
                             robot_orientation,
                             useFixedBase=True)

bullet.changeDynamics(id_robot, -1, linearDamping=0, angularDamping=0)
bullet.changeDynamics(id_robot, 0, linearDamping=0, angularDamping=0)
bullet.changeDynamics(id_robot, 1, linearDamping=0, angularDamping=0)

jointTypeNames = [
    "JOINT_REVOLUTE", "JOINT_PRISMATIC", "JOINT_SPHERICAL", "JOINT_PLANAR", "JOINT_FIXED",
    "JOINT_POINT2POINT", "JOINT_GEAR"
]

# Disable the motors for torque control:
bullet.setJointMotorControlArray(id_robot,
                                 id_revolute_joints,
                                 bullet.VELOCITY_CONTROL,
                                 forces=[0.0, 0.0])

# Target Positions:
start = 0.0
end = 1.0

steps = int((end - start) / delta_t)
t = [0] * steps
q_pos_desired = [[0.] * steps, [0.] * steps]
q_vel_desired = [[0.] * steps, [0.] * steps]
q_acc_desired = [[0.] * steps, [0.] * steps]

for s in range(steps):
  t[s] = start + s * delta_t
  q_pos_desired[0][s] = 1. / (2. * math.pi) * math.sin(2. * math.pi * t[s]) - t[s]
  q_pos_desired[1][s] = -1. / (2. * math.pi) * (math.cos(2. * math.pi * t[s]) - 1.0)

  q_vel_desired[0][s] = math.cos(2. * math.pi * t[s]) - 1.
  q_vel_desired[1][s] = math.sin(2. * math.pi * t[s])

  q_acc_desired[0][s] = -2. * math.pi * math.sin(2. * math.pi * t[s])
  q_acc_desired[1][s] = 2. * math.pi * math.cos(2. * math.pi * t[s])

q_pos = [[0.] * steps, [0.] * steps]
q_vel = [[0.] * steps, [0.] * steps]
q_tor = [[0.] * steps, [0.] * steps]

# Do Torque Control:
for i in range(len(t)):

  # Read Sensor States:
  joint_states = bullet.getJointStates(id_robot, id_revolute_joints)

  q_pos[0][i] = joint_states[0][0]
  a = joint_states[1][0]
  if (verbose):
    print("joint_states[1][0]")
    print(joint_states[1][0])
  q_pos[1][i] = a

  q_vel[0][i] = joint_states[0][1]
  q_vel[1][i] = joint_states[1][1]

  # Computing the torque from inverse dynamics:
  obj_pos = [q_pos[0][i], q_pos[1][i]]
  obj_vel = [q_vel[0][i], q_vel[1][i]]
  obj_acc = [q_acc_desired[0][i], q_acc_desired[1][i]]

  if (verbose):
    print("calculateInverseDynamics")
    print("id_robot")
    print(id_robot)
    print("obj_pos")
    print(obj_pos)
    print("obj_vel")
    print(obj_vel)
    print("obj_acc")
    print(obj_acc)

  torque = bullet.calculateInverseDynamics(id_robot, obj_pos, obj_vel, obj_acc)
  q_tor[0][i] = torque[0]
  q_tor[1][i] = torque[1]
  if (verbose):
    print("torque=")
    print(torque)

  # Set the Joint Torques:
  bullet.setJointMotorControlArray(id_robot,
                                   id_revolute_joints,
                                   bullet.TORQUE_CONTROL,
                                   forces=[torque[0], torque[1]])

  # Step Simulation
  bullet.stepSimulation()

# Plot the Position, Velocity and Acceleration:
if plot:
  figure = plt.figure(figsize=[15, 4.5])
  figure.subplots_adjust(left=0.05, bottom=0.11, right=0.97, top=0.9, wspace=0.4, hspace=0.55)

  ax_pos = figure.add_subplot(141)
  ax_pos.set_title("Joint Position")
  ax_pos.plot(t, q_pos_desired[0], '--r', lw=4, label='Desired q0')
  ax_pos.plot(t, q_pos_desired[1], '--b', lw=4, label='Desired q1')
  ax_pos.plot(t, q_pos[0], '-r', lw=1, label='Measured q0')
  ax_pos.plot(t, q_pos[1], '-b', lw=1, label='Measured q1')
  ax_pos.set_ylim(-1., 1.)
  ax_pos.legend()

  ax_vel = figure.add_subplot(142)
  ax_vel.set_title("Joint Velocity")
  ax_vel.plot(t, q_vel_desired[0], '--r', lw=4, label='Desired q0')
  ax_vel.plot(t, q_vel_desired[1], '--b', lw=4, label='Desired q1')
  ax_vel.plot(t, q_vel[0], '-r', lw=1, label='Measured q0')
  ax_vel.plot(t, q_vel[1], '-b', lw=1, label='Measured q1')
  ax_vel.set_ylim(-2., 2.)
  ax_vel.legend()

  ax_acc = figure.add_subplot(143)
  ax_acc.set_title("Joint Acceleration")
  ax_acc.plot(t, q_acc_desired[0], '--r', lw=4, label='Desired q0')
  ax_acc.plot(t, q_acc_desired[1], '--b', lw=4, label='Desired q1')
  ax_acc.set_ylim(-10., 10.)
  ax_acc.legend()

  ax_tor = figure.add_subplot(144)
  ax_tor.set_title("Executed Torque")
  ax_tor.plot(t, q_tor[0], '-r', lw=2, label='Torque q0')
  ax_tor.plot(t, q_tor[1], '-b', lw=2, label='Torque q1')
  ax_tor.set_ylim(-20., 20.)
  ax_tor.legend()

  plt.pause(0.01)

while (1):
  bullet.stepSimulation()
  time.sleep(0.01)



"""This file implements the functionalities of a minitaur using pybullet.

"""

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import collections
import copy
import math
import re

import numpy as np
from pybullet_envs.minitaur.envs import motor

INIT_POSITION = [0, 0, .2]
INIT_RACK_POSITION = [0, 0, 1]
INIT_ORIENTATION = [0, 0, 0, 1]
KNEE_CONSTRAINT_POINT_RIGHT = [0, 0.005, 0.2]
KNEE_CONSTRAINT_POINT_LEFT = [0, 0.01, 0.2]
OVERHEAT_SHUTDOWN_TORQUE = 2.45
OVERHEAT_SHUTDOWN_TIME = 1.0
LEG_POSITION = ["front_left", "back_left", "front_right", "back_right"]
MOTOR_NAMES = [
    "motor_front_leftL_joint", "motor_front_leftR_joint", "motor_back_leftL_joint",
    "motor_back_leftR_joint", "motor_front_rightL_joint", "motor_front_rightR_joint",
    "motor_back_rightL_joint", "motor_back_rightR_joint"
]
_CHASSIS_NAME_PATTERN = re.compile(r"chassis\D*center")
_MOTOR_NAME_PATTERN = re.compile(r"motor\D*joint")
_KNEE_NAME_PATTERN = re.compile(r"knee\D*")
SENSOR_NOISE_STDDEV = (0.0, 0.0, 0.0, 0.0, 0.0)
TWO_PI = 2 * math.pi


def MapToMinusPiToPi(angles):
  """Maps a list of angles to [-pi, pi].

  Args:
    angles: A list of angles in rad.
  Returns:
    A list of angle mapped to [-pi, pi].
  """
  mapped_angles = copy.deepcopy(angles)
  for i in range(len(angles)):
    mapped_angles[i] = math.fmod(angles[i], TWO_PI)
    if mapped_angles[i] >= math.pi:
      mapped_angles[i] -= TWO_PI
    elif mapped_angles[i] < -math.pi:
      mapped_angles[i] += TWO_PI
  return mapped_angles


class Minitaur(object):
  """The minitaur class that simulates a quadruped robot from Ghost Robotics.

  """

  def __init__(self,
               pybullet_client,
               urdf_root="",
               time_step=0.01,
               action_repeat=1,
               self_collision_enabled=False,
               motor_velocity_limit=np.inf,
               pd_control_enabled=False,
               accurate_motor_model_enabled=False,
               remove_default_joint_damping=False,
               motor_kp=1.0,
               motor_kd=0.02,
               pd_latency=0.0,
               control_latency=0.0,
               observation_noise_stdev=SENSOR_NOISE_STDDEV,
               torque_control_enabled=False,
               motor_overheat_protection=False,
               on_rack=False):
    """Constructs a minitaur and reset it to the initial states.

    Args:
      pybullet_client: The instance of BulletClient to manage different
        simulations.
      urdf_root: The path to the urdf folder.
      time_step: The time step of the simulation.
      action_repeat: The number of ApplyAction() for each control step.
      self_collision_enabled: Whether to enable self collision.
      motor_velocity_limit: The upper limit of the motor velocity.
      pd_control_enabled: Whether to use PD control for the motors.
      accurate_motor_model_enabled: Whether to use the accurate DC motor model.
      remove_default_joint_damping: Whether to remove the default joint damping.
      motor_kp: proportional gain for the accurate motor model.
      motor_kd: derivative gain for the accurate motor model.
      pd_latency: The latency of the observations (in seconds) used to calculate
        PD control. On the real hardware, it is the latency between the
        microcontroller and the motor controller.
      control_latency: The latency of the observations (in second) used to
        calculate action. On the real hardware, it is the latency from the motor
        controller, the microcontroller to the host (Nvidia TX2).
      observation_noise_stdev: The standard deviation of a Gaussian noise model
        for the sensor. It should be an array for separate sensors in the
        following order [motor_angle, motor_velocity, motor_torque,
        base_roll_pitch_yaw, base_angular_velocity]
      torque_control_enabled: Whether to use the torque control, if set to
        False, pose control will be used.
      motor_overheat_protection: Whether to shutdown the motor that has exerted
        large torque (OVERHEAT_SHUTDOWN_TORQUE) for an extended amount of time
        (OVERHEAT_SHUTDOWN_TIME). See ApplyAction() in minitaur.py for more
        details.
      on_rack: Whether to place the minitaur on rack. This is only used to debug
        the walking gait. In this mode, the minitaur's base is hanged midair so
        that its walking gait is clearer to visualize.
    """
    self.num_motors = 8
    self.num_legs = int(self.num_motors / 2)
    self._pybullet_client = pybullet_client
    self._action_repeat = action_repeat
    self._urdf_root = urdf_root
    self._self_collision_enabled = self_collision_enabled
    self._motor_velocity_limit = motor_velocity_limit
    self._pd_control_enabled = pd_control_enabled
    self._motor_direction = [-1, -1, -1, -1, 1, 1, 1, 1]
    self._observed_motor_torques = np.zeros(self.num_motors)
    self._applied_motor_torques = np.zeros(self.num_motors)
    self._max_force = 3.5
    self._pd_latency = pd_latency
    self._control_latency = control_latency
    self._observation_noise_stdev = observation_noise_stdev
    self._accurate_motor_model_enabled = accurate_motor_model_enabled
    self._remove_default_joint_damping = remove_default_joint_damping
    self._observation_history = collections.deque(maxlen=100)
    self._control_observation = []
    self._chassis_link_ids = [-1]
    self._leg_link_ids = []
    self._motor_link_ids = []
    self._foot_link_ids = []
    self._torque_control_enabled = torque_control_enabled
    self._motor_overheat_protection = motor_overheat_protection
    self._on_rack = on_rack
    if self._accurate_motor_model_enabled:
      self._kp = motor_kp
      self._kd = motor_kd
      self._motor_model = motor.MotorModel(torque_control_enabled=self._torque_control_enabled,
                                           kp=self._kp,
                                           kd=self._kd)
    elif self._pd_control_enabled:
      self._kp = 8
      self._kd = 0.3
    else:
      self._kp = 1
      self._kd = 1
    self.time_step = time_step
    self._step_counter = 0
    # reset_time=-1.0 means skipping the reset motion.
    # See Reset for more details.
    self.Reset(reset_time=-1.0)

  def GetTimeSinceReset(self):
    return self._step_counter * self.time_step

  def Step(self, action):
    for _ in range(self._action_repeat):
      self.ApplyAction(action)
      self._pybullet_client.stepSimulation()
      self.ReceiveObservation()
      self._step_counter += 1

  def Terminate(self):
    pass

  def _RecordMassInfoFromURDF(self):
    self._base_mass_urdf = []
    for chassis_id in self._chassis_link_ids:
      self._base_mass_urdf.append(
          self._pybullet_client.getDynamicsInfo(self.quadruped, chassis_id)[0])
    self._leg_masses_urdf = []
    for leg_id in self._leg_link_ids:
      self._leg_masses_urdf.append(
          self._pybullet_client.getDynamicsInfo(self.quadruped, leg_id)[0])
    for motor_id in self._motor_link_ids:
      self._leg_masses_urdf.append(
          self._pybullet_client.getDynamicsInfo(self.quadruped, motor_id)[0])

  def _RecordInertiaInfoFromURDF(self):
    """Record the inertia of each body from URDF file."""
    self._link_urdf = []
    num_bodies = self._pybullet_client.getNumJoints(self.quadruped)
    for body_id in range(-1, num_bodies):  # -1 is for the base link.
      inertia = self._pybullet_client.getDynamicsInfo(self.quadruped, body_id)[2]
      self._link_urdf.append(inertia)
    # We need to use id+1 to index self._link_urdf because it has the base
    # (index = -1) at the first element.
    self._base_inertia_urdf = [
        self._link_urdf[chassis_id + 1] for chassis_id in self._chassis_link_ids
    ]
    self._leg_inertia_urdf = [self._link_urdf[leg_id + 1] for leg_id in self._leg_link_ids]
    self._leg_inertia_urdf.extend(
        [self._link_urdf[motor_id + 1] for motor_id in self._motor_link_ids])

  def _BuildJointNameToIdDict(self):
    num_joints = self._pybullet_client.getNumJoints(self.quadruped)
    self._joint_name_to_id = {}
    for i in range(num_joints):
      joint_info = self._pybullet_client.getJointInfo(self.quadruped, i)
      self._joint_name_to_id[joint_info[1].decode("UTF-8")] = joint_info[0]

  def _BuildUrdfIds(self):
    """Build the link Ids from its name in the URDF file."""
    num_joints = self._pybullet_client.getNumJoints(self.quadruped)
    self._chassis_link_ids = [-1]
    # the self._leg_link_ids include both the upper and lower links of the leg.
    self._leg_link_ids = []
    self._motor_link_ids = []
    self._foot_link_ids = []
    for i in range(num_joints):
      joint_info = self._pybullet_client.getJointInfo(self.quadruped, i)
      joint_name = joint_info[1].decode("UTF-8")
      joint_id = self._joint_name_to_id[joint_name]
      if _CHASSIS_NAME_PATTERN.match(joint_name):
        self._chassis_link_ids.append(joint_id)
      elif _MOTOR_NAME_PATTERN.match(joint_name):
        self._motor_link_ids.append(joint_id)
      elif _KNEE_NAME_PATTERN.match(joint_name):
        self._foot_link_ids.append(joint_id)
      else:
        self._leg_link_ids.append(joint_id)
    self._leg_link_ids.extend(self._foot_link_ids)
    self._chassis_link_ids.sort()
    self._motor_link_ids.sort()
    self._foot_link_ids.sort()
    self._leg_link_ids.sort()

  def _RemoveDefaultJointDamping(self):
    num_joints = self._pybullet_client.getNumJoints(self.quadruped)
    for i in range(num_joints):
      joint_info = self._pybullet_client.getJointInfo(self.quadruped, i)
      self._pybullet_client.changeDynamics(joint_info[0], -1, linearDamping=0, angularDamping=0)

  def _BuildMotorIdList(self):
    self._motor_id_list = [self._joint_name_to_id[motor_name] for motor_name in MOTOR_NAMES]

  def IsObservationValid(self):
    """Whether the observation is valid for the current time step.

    In simulation, observations are always valid. In real hardware, it may not
    be valid from time to time when communication error happens between the
    Nvidia TX2 and the microcontroller.

    Returns:
      Whether the observation is valid for the current time step.
    """
    return True

  def Reset(self, reload_urdf=True, default_motor_angles=None, reset_time=3.0):
    """Reset the minitaur to its initial states.

    Args:
      reload_urdf: Whether to reload the urdf file. If not, Reset() just place
        the minitaur back to its starting position.
      default_motor_angles: The default motor angles. If it is None, minitaur
        will hold a default pose (motor angle math.pi / 2) for 100 steps. In
        torque control mode, the phase of holding the default pose is skipped.
      reset_time: The duration (in seconds) to hold the default motor angles. If
        reset_time <= 0 or in torque control mode, the phase of holding the
        default pose is skipped.
    """
    if self._on_rack:
      init_position = INIT_RACK_POSITION
    else:
      init_position = INIT_POSITION
    if reload_urdf:
      if self._self_collision_enabled:
        self.quadruped = self._pybullet_client.loadURDF(
            "%s/quadruped/minitaur.urdf" % self._urdf_root,
            init_position,
            useFixedBase=self._on_rack,
            flags=self._pybullet_client.URDF_USE_SELF_COLLISION)
      else:
        self.quadruped = self._pybullet_client.loadURDF("%s/quadruped/minitaur.urdf" %
                                                        self._urdf_root,
                                                        init_position,
                                                        useFixedBase=self._on_rack)
      self._BuildJointNameToIdDict()
      self._BuildUrdfIds()
      if self._remove_default_joint_damping:
        self._RemoveDefaultJointDamping()
      self._BuildMotorIdList()
      self._RecordMassInfoFromURDF()
      self._RecordInertiaInfoFromURDF()
      self.ResetPose(add_constraint=True)
    else:
      self._pybullet_client.resetBasePositionAndOrientation(self.quadruped, init_position,
                                                            INIT_ORIENTATION)
      self._pybullet_client.resetBaseVelocity(self.quadruped, [0, 0, 0], [0, 0, 0])
      self.ResetPose(add_constraint=False)
    self._overheat_counter = np.zeros(self.num_motors)
    self._motor_enabled_list = [True] * self.num_motors
    self._step_counter = 0

    # Perform reset motion within reset_duration if in position control mode.
    # Nothing is performed if in torque control mode for now.
    # TODO(jietan): Add reset motion when the torque control is fully supported.
    self._observation_history.clear()
    if not self._torque_control_enabled and reset_time > 0.0:
      self.ReceiveObservation()
      for _ in range(100):
        self.ApplyAction([math.pi / 2] * self.num_motors)
        self._pybullet_client.stepSimulation()
        self.ReceiveObservation()
      if default_motor_angles is not None:
        num_steps_to_reset = int(reset_time / self.time_step)
        for _ in range(num_steps_to_reset):
          self.ApplyAction(default_motor_angles)
          self._pybullet_client.stepSimulation()
          self.ReceiveObservation()
    self.ReceiveObservation()

  def _SetMotorTorqueById(self, motor_id, torque):
    self._pybullet_client.setJointMotorControl2(bodyIndex=self.quadruped,
                                                jointIndex=motor_id,
                                                controlMode=self._pybullet_client.TORQUE_CONTROL,
                                                force=torque)

  def _SetDesiredMotorAngleById(self, motor_id, desired_angle):
    self._pybullet_client.setJointMotorControl2(bodyIndex=self.quadruped,
                                                jointIndex=motor_id,
                                                controlMode=self._pybullet_client.POSITION_CONTROL,
                                                targetPosition=desired_angle,
                                                positionGain=self._kp,
                                                velocityGain=self._kd,
                                                force=self._max_force)

  def _SetDesiredMotorAngleByName(self, motor_name, desired_angle):
    self._SetDesiredMotorAngleById(self._joint_name_to_id[motor_name], desired_angle)

  def ResetPose(self, add_constraint):
    """Reset the pose of the minitaur.

    Args:
      add_constraint: Whether to add a constraint at the joints of two feet.
    """
    for i in range(self.num_legs):
      self._ResetPoseForLeg(i, add_constraint)

  def _ResetPoseForLeg(self, leg_id, add_constraint):
    """Reset the initial pose for the leg.

    Args:
      leg_id: It should be 0, 1, 2, or 3, which represents the leg at
        front_left, back_left, front_right and back_right.
      add_constraint: Whether to add a constraint at the joints of two feet.
    """
    knee_friction_force = 0
    half_pi = math.pi / 2.0
    knee_angle = -2.1834

    leg_position = LEG_POSITION[leg_id]
    self._pybullet_client.resetJointState(self.quadruped,
                                          self._joint_name_to_id["motor_" + leg_position +
                                                                 "L_joint"],
                                          self._motor_direction[2 * leg_id] * half_pi,
                                          targetVelocity=0)
    self._pybullet_client.resetJointState(self.quadruped,
                                          self._joint_name_to_id["knee_" + leg_position +
                                                                 "L_link"],
                                          self._motor_direction[2 * leg_id] * knee_angle,
                                          targetVelocity=0)
    self._pybullet_client.resetJointState(self.quadruped,
                                          self._joint_name_to_id["motor_" + leg_position +
                                                                 "R_joint"],
                                          self._motor_direction[2 * leg_id + 1] * half_pi,
                                          targetVelocity=0)
    self._pybullet_client.resetJointState(self.quadruped,
                                          self._joint_name_to_id["knee_" + leg_position +
                                                                 "R_link"],
                                          self._motor_direction[2 * leg_id + 1] * knee_angle,
                                          targetVelocity=0)
    if add_constraint:
      self._pybullet_client.createConstraint(
          self.quadruped, self._joint_name_to_id["knee_" + leg_position + "R_link"],
          self.quadruped, self._joint_name_to_id["knee_" + leg_position + "L_link"],
          self._pybullet_client.JOINT_POINT2POINT, [0, 0, 0], KNEE_CONSTRAINT_POINT_RIGHT,
          KNEE_CONSTRAINT_POINT_LEFT)

    if self._accurate_motor_model_enabled or self._pd_control_enabled:
      # Disable the default motor in pybullet.
      self._pybullet_client.setJointMotorControl2(
          bodyIndex=self.quadruped,
          jointIndex=(self._joint_name_to_id["motor_" + leg_position + "L_joint"]),
          controlMode=self._pybullet_client.VELOCITY_CONTROL,
          targetVelocity=0,
          force=knee_friction_force)
      self._pybullet_client.setJointMotorControl2(
          bodyIndex=self.quadruped,
          jointIndex=(self._joint_name_to_id["motor_" + leg_position + "R_joint"]),
          controlMode=self._pybullet_client.VELOCITY_CONTROL,
          targetVelocity=0,
          force=knee_friction_force)

    else:
      self._SetDesiredMotorAngleByName("motor_" + leg_position + "L_joint",
                                       self._motor_direction[2 * leg_id] * half_pi)
      self._SetDesiredMotorAngleByName("motor_" + leg_position + "R_joint",
                                       self._motor_direction[2 * leg_id + 1] * half_pi)

    self._pybullet_client.setJointMotorControl2(
        bodyIndex=self.quadruped,
        jointIndex=(self._joint_name_to_id["knee_" + leg_position + "L_link"]),
        controlMode=self._pybullet_client.VELOCITY_CONTROL,
        targetVelocity=0,
        force=knee_friction_force)
    self._pybullet_client.setJointMotorControl2(
        bodyIndex=self.quadruped,
        jointIndex=(self._joint_name_to_id["knee_" + leg_position + "R_link"]),
        controlMode=self._pybullet_client.VELOCITY_CONTROL,
        targetVelocity=0,
        force=knee_friction_force)

  def GetBasePosition(self):
    """Get the position of minitaur's base.

    Returns:
      The position of minitaur's base.
    """
    position, _ = (self._pybullet_client.getBasePositionAndOrientation(self.quadruped))
    return position

  def GetTrueBaseRollPitchYaw(self):
    """Get minitaur's base orientation in euler angle in the world frame.

    Returns:
      A tuple (roll, pitch, yaw) of the base in world frame.
    """
    orientation = self.GetTrueBaseOrientation()
    roll_pitch_yaw = self._pybullet_client.getEulerFromQuaternion(orientation)
    return np.asarray(roll_pitch_yaw)

  def GetBaseRollPitchYaw(self):
    """Get minitaur's base orientation in euler angle in the world frame.

    This function mimicks the noisy sensor reading and adds latency.
    Returns:
      A tuple (roll, pitch, yaw) of the base in world frame polluted by noise
      and latency.
    """
    delayed_orientation = np.array(
        self._control_observation[3 * self.num_motors:3 * self.num_motors + 4])
    delayed_roll_pitch_yaw = self._pybullet_client.getEulerFromQuaternion(delayed_orientation)
    roll_pitch_yaw = self._AddSensorNoise(np.array(delayed_roll_pitch_yaw),
                                          self._observation_noise_stdev[3])
    return roll_pitch_yaw

  def GetTrueMotorAngles(self):
    """Gets the eight motor angles at the current moment, mapped to [-pi, pi].

    Returns:
      Motor angles, mapped to [-pi, pi].
    """
    motor_angles = [
        self._pybullet_client.getJointState(self.quadruped, motor_id)[0]
        for motor_id in self._motor_id_list
    ]
    motor_angles = np.multiply(motor_angles, self._motor_direction)
    return motor_angles

  def GetMotorAngles(self):
    """Gets the eight motor angles.

    This function mimicks the noisy sensor reading and adds latency. The motor
    angles that are delayed, noise polluted, and mapped to [-pi, pi].

    Returns:
      Motor angles polluted by noise and latency, mapped to [-pi, pi].
    """
    motor_angles = self._AddSensorNoise(np.array(self._control_observation[0:self.num_motors]),
                                        self._observation_noise_stdev[0])
    return MapToMinusPiToPi(motor_angles)

  def GetTrueMotorVelocities(self):
    """Get the velocity of all eight motors.

    Returns:
      Velocities of all eight motors.
    """
    motor_velocities = [
        self._pybullet_client.getJointState(self.quadruped, motor_id)[1]
        for motor_id in self._motor_id_list
    ]
    motor_velocities = np.multiply(motor_velocities, self._motor_direction)
    return motor_velocities

  def GetMotorVelocities(self):
    """Get the velocity of all eight motors.

    This function mimicks the noisy sensor reading and adds latency.
    Returns:
      Velocities of all eight motors polluted by noise and latency.
    """
    return self._AddSensorNoise(
        np.array(self._control_observation[self.num_motors:2 * self.num_motors]),
        self._observation_noise_stdev[1])

  def GetTrueMotorTorques(self):
    """Get the amount of torque the motors are exerting.

    Returns:
      Motor torques of all eight motors.
    """
    if self._accurate_motor_model_enabled or self._pd_control_enabled:
      return self._observed_motor_torques
    else:
      motor_torques = [
          self._pybullet_client.getJointState(self.quadruped, motor_id)[3]
          for motor_id in self._motor_id_list
      ]
      motor_torques = np.multiply(motor_torques, self._motor_direction)
    return motor_torques

  def GetMotorTorques(self):
    """Get the amount of torque the motors are exerting.

    This function mimicks the noisy sensor reading and adds latency.
    Returns:
      Motor torques of all eight motors polluted by noise and latency.
    """
    return self._AddSensorNoise(
        np.array(self._control_observation[2 * self.num_motors:3 * self.num_motors]),
        self._observation_noise_stdev[2])

  def GetTrueBaseOrientation(self):
    """Get the orientation of minitaur's base, represented as quaternion.

    Returns:
      The orientation of minitaur's base.
    """
    _, orientation = (self._pybullet_client.getBasePositionAndOrientation(self.quadruped))
    return orientation

  def GetBaseOrientation(self):
    """Get the orientation of minitaur's base, represented as quaternion.

    This function mimicks the noisy sensor reading and adds latency.
    Returns:
      The orientation of minitaur's base polluted by noise and latency.
    """
    return self._pybullet_client.getQuaternionFromEuler(self.GetBaseRollPitchYaw())

  def GetTrueBaseRollPitchYawRate(self):
    """Get the rate of orientation change of the minitaur's base in euler angle.

    Returns:
      rate of (roll, pitch, yaw) change of the minitaur's base.
    """
    vel = self._pybullet_client.getBaseVelocity(self.quadruped)
    return np.asarray([vel[1][0], vel[1][1], vel[1][2]])

  def GetBaseRollPitchYawRate(self):
    """Get the rate of orientation change of the minitaur's base in euler angle.

    This function mimicks the noisy sensor reading and adds latency.
    Returns:
      rate of (roll, pitch, yaw) change of the minitaur's base polluted by noise
      and latency.
    """
    return self._AddSensorNoise(
        np.array(self._control_observation[3 * self.num_motors + 4:3 * self.num_motors + 7]),
        self._observation_noise_stdev[4])

  def GetActionDimension(self):
    """Get the length of the action list.

    Returns:
      The length of the action list.
    """
    return self.num_motors

  def ApplyAction(self, motor_commands, motor_kps=None, motor_kds=None):
    """Set the desired motor angles to the motors of the minitaur.

    The desired motor angles are clipped based on the maximum allowed velocity.
    If the pd_control_enabled is True, a torque is calculated according to
    the difference between current and desired joint angle, as well as the joint
    velocity. This torque is exerted to the motor. For more information about
    PD control, please refer to: https://en.wikipedia.org/wiki/PID_controller.

    Args:
      motor_commands: The eight desired motor angles.
      motor_kps: Proportional gains for the motor model. If not provided, it
        uses the default kp of the minitaur for all the motors.
      motor_kds: Derivative gains for the motor model. If not provided, it
        uses the default kd of the minitaur for all the motors.
    """
    if self._motor_velocity_limit < np.inf:
      current_motor_angle = self.GetTrueMotorAngles()
      motor_commands_max = (current_motor_angle + self.time_step * self._motor_velocity_limit)
      motor_commands_min = (current_motor_angle - self.time_step * self._motor_velocity_limit)
      motor_commands = np.clip(motor_commands, motor_commands_min, motor_commands_max)
    # Set the kp and kd for all the motors if not provided as an argument.
    if motor_kps is None:
      motor_kps = np.full(8, self._kp)
    if motor_kds is None:
      motor_kds = np.full(8, self._kd)

    if self._accurate_motor_model_enabled or self._pd_control_enabled:
      q, qdot = self._GetPDObservation()
      qdot_true = self.GetTrueMotorVelocities()
      if self._accurate_motor_model_enabled:
        actual_torque, observed_torque = self._motor_model.convert_to_torque(
            motor_commands, q, qdot, qdot_true, motor_kps, motor_kds)
        if self._motor_overheat_protection:
          for i in range(self.num_motors):
            if abs(actual_torque[i]) > OVERHEAT_SHUTDOWN_TORQUE:
              self._overheat_counter[i] += 1
            else:
              self._overheat_counter[i] = 0
            if (self._overheat_counter[i] > OVERHEAT_SHUTDOWN_TIME / self.time_step):
              self._motor_enabled_list[i] = False

        # The torque is already in the observation space because we use
        # GetMotorAngles and GetMotorVelocities.
        self._observed_motor_torques = observed_torque

        # Transform into the motor space when applying the torque.
        self._applied_motor_torque = np.multiply(actual_torque, self._motor_direction)

        for motor_id, motor_torque, motor_enabled in zip(self._motor_id_list,
                                                         self._applied_motor_torque,
                                                         self._motor_enabled_list):
          if motor_enabled:
            self._SetMotorTorqueById(motor_id, motor_torque)
          else:
            self._SetMotorTorqueById(motor_id, 0)
      else:
        torque_commands = -1 * motor_kps * (q - motor_commands) - motor_kds * qdot

        # The torque is already in the observation space because we use
        # GetMotorAngles and GetMotorVelocities.
        self._observed_motor_torques = torque_commands

        # Transform into the motor space when applying the torque.
        self._applied_motor_torques = np.multiply(self._observed_motor_torques,
                                                  self._motor_direction)

        for motor_id, motor_torque in zip(self._motor_id_list, self._applied_motor_torques):
          self._SetMotorTorqueById(motor_id, motor_torque)
    else:
      motor_commands_with_direction = np.multiply(motor_commands, self._motor_direction)
      for motor_id, motor_command_with_direction in zip(self._motor_id_list,
                                                        motor_commands_with_direction):
        self._SetDesiredMotorAngleById(motor_id, motor_command_with_direction)

  def ConvertFromLegModel(self, actions):
    """Convert the actions that use leg model to the real motor actions.

    Args:
      actions: The theta, phi of the leg model.
    Returns:
      The eight desired motor angles that can be used in ApplyActions().
    """
    motor_angle = copy.deepcopy(actions)
    scale_for_singularity = 1
    offset_for_singularity = 1.5
    half_num_motors = int(self.num_motors / 2)
    quater_pi = math.pi / 4
    for i in range(self.num_motors):
      action_idx = int(i // 2)
      forward_backward_component = (
          -scale_for_singularity * quater_pi *
          (actions[action_idx + half_num_motors] + offset_for_singularity))
      extension_component = (-1)**i * quater_pi * actions[action_idx]
      if i >= half_num_motors:
        extension_component = -extension_component
      motor_angle[i] = (math.pi + forward_backward_component + extension_component)
    return motor_angle

  def GetBaseMassesFromURDF(self):
    """Get the mass of the base from the URDF file."""
    return self._base_mass_urdf

  def GetBaseInertiasFromURDF(self):
    """Get the inertia of the base from the URDF file."""
    return self._base_inertia_urdf

  def GetLegMassesFromURDF(self):
    """Get the mass of the legs from the URDF file."""
    return self._leg_masses_urdf

  def GetLegInertiasFromURDF(self):
    """Get the inertia of the legs from the URDF file."""
    return self._leg_inertia_urdf

  def SetBaseMasses(self, base_mass):
    """Set the mass of minitaur's base.

    Args:
      base_mass: A list of masses of each body link in CHASIS_LINK_IDS. The
        length of this list should be the same as the length of CHASIS_LINK_IDS.
    Raises:
      ValueError: It is raised when the length of base_mass is not the same as
        the length of self._chassis_link_ids.
    """
    if len(base_mass) != len(self._chassis_link_ids):
      raise ValueError("The length of base_mass {} and self._chassis_link_ids {} are not "
                       "the same.".format(len(base_mass), len(self._chassis_link_ids)))
    for chassis_id, chassis_mass in zip(self._chassis_link_ids, base_mass):
      self._pybullet_client.changeDynamics(self.quadruped, chassis_id, mass=chassis_mass)

  def SetLegMasses(self, leg_masses):
    """Set the mass of the legs.

    A leg includes leg_link and motor. 4 legs contain 16 links (4 links each)
    and 8 motors. First 16 numbers correspond to link masses, last 8 correspond
    to motor masses (24 total).

    Args:
      leg_masses: The leg and motor masses for all the leg links and motors.

    Raises:
      ValueError: It is raised when the length of masses is not equal to number
        of links + motors.
    """
    if len(leg_masses) != len(self._leg_link_ids) + len(self._motor_link_ids):
      raise ValueError("The number of values passed to SetLegMasses are "
                       "different than number of leg links and motors.")
    for leg_id, leg_mass in zip(self._leg_link_ids, leg_masses):
      self._pybullet_client.changeDynamics(self.quadruped, leg_id, mass=leg_mass)
    motor_masses = leg_masses[len(self._leg_link_ids):]
    for link_id, motor_mass in zip(self._motor_link_ids, motor_masses):
      self._pybullet_client.changeDynamics(self.quadruped, link_id, mass=motor_mass)

  def SetBaseInertias(self, base_inertias):
    """Set the inertias of minitaur's base.

    Args:
      base_inertias: A list of inertias of each body link in CHASIS_LINK_IDS.
        The length of this list should be the same as the length of
        CHASIS_LINK_IDS.
    Raises:
      ValueError: It is raised when the length of base_inertias is not the same
        as the length of self._chassis_link_ids and base_inertias contains
        negative values.
    """
    if len(base_inertias) != len(self._chassis_link_ids):
      raise ValueError("The length of base_inertias {} and self._chassis_link_ids {} are "
                       "not the same.".format(len(base_inertias), len(self._chassis_link_ids)))
    for chassis_id, chassis_inertia in zip(self._chassis_link_ids, base_inertias):
      for inertia_value in chassis_inertia:
        if (np.asarray(inertia_value) < 0).any():
          raise ValueError("Values in inertia matrix should be non-negative.")
      self._pybullet_client.changeDynamics(self.quadruped,
                                           chassis_id,
                                           localInertiaDiagonal=chassis_inertia)

  def SetLegInertias(self, leg_inertias):
    """Set the inertias of the legs.

    A leg includes leg_link and motor. 4 legs contain 16 links (4 links each)
    and 8 motors. First 16 numbers correspond to link inertia, last 8 correspond
    to motor inertia (24 total).

    Args:
      leg_inertias: The leg and motor inertias for all the leg links and motors.

    Raises:
      ValueError: It is raised when the length of inertias is not equal to
      the number of links + motors or leg_inertias contains negative values.
    """

    if len(leg_inertias) != len(self._leg_link_ids) + len(self._motor_link_ids):
      raise ValueError("The number of values passed to SetLegMasses are "
                       "different than number of leg links and motors.")
    for leg_id, leg_inertia in zip(self._leg_link_ids, leg_inertias):
      for inertia_value in leg_inertias:
        if (np.asarray(inertia_value) < 0).any():
          raise ValueError("Values in inertia matrix should be non-negative.")
      self._pybullet_client.changeDynamics(self.quadruped,
                                           leg_id,
                                           localInertiaDiagonal=leg_inertia)

    motor_inertias = leg_inertias[len(self._leg_link_ids):]
    for link_id, motor_inertia in zip(self._motor_link_ids, motor_inertias):
      for inertia_value in motor_inertias:
        if (np.asarray(inertia_value) < 0).any():
          raise ValueError("Values in inertia matrix should be non-negative.")
      self._pybullet_client.changeDynamics(self.quadruped,
                                           link_id,
                                           localInertiaDiagonal=motor_inertia)

  def SetFootFriction(self, foot_friction):
    """Set the lateral friction of the feet.

    Args:
      foot_friction: The lateral friction coefficient of the foot. This value is
        shared by all four feet.
    """
    for link_id in self._foot_link_ids:
      self._pybullet_client.changeDynamics(self.quadruped, link_id, lateralFriction=foot_friction)

  # TODO(b/73748980): Add more API's to set other contact parameters.
  def SetFootRestitution(self, foot_restitution):
    """Set the coefficient of restitution at the feet.

    Args:
      foot_restitution: The coefficient of restitution (bounciness) of the feet.
        This value is shared by all four feet.
    """
    for link_id in self._foot_link_ids:
      self._pybullet_client.changeDynamics(self.quadruped, link_id, restitution=foot_restitution)

  def SetJointFriction(self, joint_frictions):
    for knee_joint_id, friction in zip(self._foot_link_ids, joint_frictions):
      self._pybullet_client.setJointMotorControl2(
          bodyIndex=self.quadruped,
          jointIndex=knee_joint_id,
          controlMode=self._pybullet_client.VELOCITY_CONTROL,
          targetVelocity=0,
          force=friction)

  def GetNumKneeJoints(self):
    return len(self._foot_link_ids)

  def SetBatteryVoltage(self, voltage):
    if self._accurate_motor_model_enabled:
      self._motor_model.set_voltage(voltage)

  def SetMotorViscousDamping(self, viscous_damping):
    if self._accurate_motor_model_enabled:
      self._motor_model.set_viscous_damping(viscous_damping)

  def GetTrueObservation(self):
    observation = []
    observation.extend(self.GetTrueMotorAngles())
    observation.extend(self.GetTrueMotorVelocities())
    observation.extend(self.GetTrueMotorTorques())
    observation.extend(self.GetTrueBaseOrientation())
    observation.extend(self.GetTrueBaseRollPitchYawRate())
    return observation

  def ReceiveObservation(self):
    """Receive the observation from sensors.

    This function is called once per step. The observations are only updated
    when this function is called.
    """
    self._observation_history.appendleft(self.GetTrueObservation())
    self._control_observation = self._GetControlObservation()

  def _GetDelayedObservation(self, latency):
    """Get observation that is delayed by the amount specified in latency.

    Args:
      latency: The latency (in seconds) of the delayed observation.
    Returns:
      observation: The observation which was actually latency seconds ago.
    """
    if latency <= 0 or len(self._observation_history) == 1:
      observation = self._observation_history[0]
    else:
      n_steps_ago = int(latency / self.time_step)
      if n_steps_ago + 1 >= len(self._observation_history):
        return self._observation_history[-1]
      remaining_latency = latency - n_steps_ago * self.time_step
      blend_alpha = remaining_latency / self.time_step
      observation = ((1.0 - blend_alpha) * np.array(self._observation_history[n_steps_ago]) +
                     blend_alpha * np.array(self._observation_history[n_steps_ago + 1]))
    return observation

  def _GetPDObservation(self):
    pd_delayed_observation = self._GetDelayedObservation(self._pd_latency)
    q = pd_delayed_observation[0:self.num_motors]
    qdot = pd_delayed_observation[self.num_motors:2 * self.num_motors]
    return (np.array(q), np.array(qdot))

  def _GetControlObservation(self):
    control_delayed_observation = self._GetDelayedObservation(self._control_latency)
    return control_delayed_observation

  def _AddSensorNoise(self, sensor_values, noise_stdev):
    if noise_stdev <= 0:
      return sensor_values
    observation = sensor_values + np.random.normal(scale=noise_stdev, size=sensor_values.shape)
    return observation

  def SetControlLatency(self, latency):
    """Set the latency of the control loop.

    It measures the duration between sending an action from Nvidia TX2 and
    receiving the observation from microcontroller.

    Args:
      latency: The latency (in seconds) of the control loop.
    """
    self._control_latency = latency

  def GetControlLatency(self):
    """Get the control latency.

    Returns:
      The latency (in seconds) between when the motor command is sent and when
        the sensor measurements are reported back to the controller.
    """
    return self._control_latency

  def SetMotorGains(self, kp, kd):
    """Set the gains of all motors.

    These gains are PD gains for motor positional control. kp is the
    proportional gain and kd is the derivative gain.

    Args:
      kp: proportional gain of the motors.
      kd: derivative gain of the motors.
    """
    self._kp = kp
    self._kd = kd
    if self._accurate_motor_model_enabled:
      self._motor_model.set_motor_gains(kp, kd)

  def GetMotorGains(self):
    """Get the gains of the motor.

    Returns:
      The proportional gain.
      The derivative gain.
    """
    return self._kp, self._kd

  def SetMotorStrengthRatio(self, ratio):
    """Set the strength of all motors relative to the default value.

    Args:
      ratio: The relative strength. A scalar range from 0.0 to 1.0.
    """
    if self._accurate_motor_model_enabled:
      self._motor_model.set_strength_ratios([ratio] * self.num_motors)

  def SetMotorStrengthRatios(self, ratios):
    """Set the strength of each motor relative to the default value.

    Args:
      ratios: The relative strength. A numpy array ranging from 0.0 to 1.0.
    """
    if self._accurate_motor_model_enabled:
      self._motor_model.set_strength_ratios(ratios)

  def SetTimeSteps(self, action_repeat, simulation_step):
    """Set the time steps of the control and simulation.

    Args:
      action_repeat: The number of simulation steps that the same action is
        repeated.
      simulation_step: The simulation time step.
    """
    self.time_step = simulation_step
    self._action_repeat = action_repeat

  @property
  def chassis_link_ids(self):
    return self._chassis_link_ids



# Lint as: python3
"""A sensor that returns the last action(s) sent to the environment."""

from typing import Any, Dict, Sequence, Text, Type, Tuple, Union
import gin
import gym
import numpy as np

from pybullet_envs.minitaur.envs_v2.sensors import sensor
from pybullet_envs.minitaur.envs_v2.sensors import space_utils


@gin.configurable
class LastActionSensor(sensor.Sensor):
  """A sensor that reports the last action taken."""

  def __init__(self,
               name: Text = "LastAction",
               dtype: Type[Any] = np.float64,
               sensor_latency: Union[float, Sequence[float]] = 0):
    """Constructs LastActionSensor.

    We do not provide a robot instance during __init__, as robot instances may
    be reloaded/recreated during the simulation.

    Args:
      name: the name of the sensor
      dtype: data type of sensor value.
      sensor_latency: There are two ways to use this expected sensor latency.
        For both methods, the latency should be in the same unit as the sensor
        data timestamp. 1. As a single float number, the observation will be a
        1D array. For real robots, this should be set to 0.0. 2. As a array of
        floats, the observation will be a 2D array based on how long the history
        need to be. Thus, [0.0, 0.1, 0.2] is a history length of 3.
    """
    super().__init__(name=name,
                     sensor_latency=sensor_latency,
                     # We generally don't interpolate actions.
                     interpolator_fn=sensor.older_obs_blender)

    self._dtype = dtype
    self._env = None

  def on_reset(self, env: gym.Env):
    """From the callback, the sensor remembers the environment.

    Args:
      env: the environment who invokes this callback function.
    """
    # Constructs the observation space using the env's action space.
    self._observation_space = self._stack_space(
        env.action_space, dtype=self._dtype)

    # Call the super class methods to initialize the buffers
    super().on_reset(env)

  def _get_original_observation(
      self) -> Tuple[float, Union[Dict[Text, np.ndarray], np.ndarray]]:
    return self._env.get_time(), space_utils.action_astype(
        self._env.last_action, self._dtype)



"""The imagery client to connect to the camera job."""

from typing import Any, Dict, Sequence, Text
import gin

from pybullet_envs.minitaur.fw_bridge import worker_builder
from pybullet_envs.minitaur.vision import imagery_pb2
from pybullet_envs.minitaur.vision import imagery_utils
from google3.third_party.fluxworks.core.fluxworks.python.genericutil_py import fwassert
from google3.third_party.fluxworks.core.fluxworks.python.genericutil_py import timeutil

_RPC_TIMEOUT = 1 * timeutil.TimeUtil.SEC

_URI_START_CAPTURE = "fwuri://VisionJob/StartCapture"
_URI_STOP_CAPTURE = "fwuri://VisionJob/StopCapture"
_URI_GET_FRAME = "fwuri://VisionJob/GetFrame"


@gin.configurable
class ImageryClient(object):
  """Sends commands and receives states from cameras."""

  def __init__(
      self,
      fw_worker=None,
      rpc_timeout_sec=_RPC_TIMEOUT,
      ip_address=None,
      port=None,
      async_mode=False,
      start_capture_uri: Text = _URI_START_CAPTURE,
      stop_capture_uri: Text = _URI_STOP_CAPTURE,
      get_frame_uri: Text = _URI_GET_FRAME,
  ):
    """Initializes the client.

    Args:
      fw_worker: A FluxWorks worker instance.
      rpc_timeout_sec: The timeout for any RPC calls from this client.
      ip_address: The ip address of the camera/vision process. If vision job is
        also instantiated in the same FluxWorks worker, both ip address and port
        are not needed.
      port: The port of the camera/vision process.
      async_mode: Whether the RPC calls in this client are async or synchronous.
        Aync mode is only required when you have multiple workers communicating
        with each other in the same Python process. If worker A is calling
        worker B's RPC, worker B's RPC is trying to get GIL from it's thread but
        caller (worker A) already holds the GIL, and this will cause a dead lock
        if worker A's calls are synchronous. If worker A is calling its own RPC,
        the same GIL can be used so there is no dead lock, and there is no need
        for async mode. Async mode will require context switching and thus is a
        bit slower.
      start_capture_uri: The FluxWorks URI to start camera capture.
      stop_capture_uri: The FluxWorks URI to stop camera capture.
      get_frame_uri: The FluxWorks URI to get camera frames.
    """
    self._rpc_timeout_sec = rpc_timeout_sec
    if fw_worker is None:
      fw_worker = worker_builder.GetDefaultWorker()
    self._worker = fw_worker

    # TODO(tingnan): Use a single address and split the string for FW.
    if ip_address is not None:
      self._worker.ConnectToWorker(ip_address, port)

    self._async_mode = async_mode
    self._start_capture_uri = start_capture_uri
    self._stop_capture_uri = stop_capture_uri
    self._get_frame_uri = get_frame_uri

  def _convert_camera_frame_to_image_dict(
      self, camera_frame: imagery_pb2.CameraFrame):
    """Converts the camera frame to an image dictionary."""
    # Each camera frame might contain multiple image channels, such as rgb and
    # depth.
    images = {}
    for image_name, image_proto in camera_frame.images.items():
      image_array = imagery_utils.convert_image_to_array(image_proto)
      images[image_name] = image_array
    return images

  def start_capture(self, run_id: Text = "vision"):
    """Starts the camera capture session.

    Args:
      run_id: The capture session id. This id will determine the name of the
        image logs' sub-direcotry.
    """
    capture_request = imagery_pb2.CaptureRequest()
    capture_request.run_id = run_id
    fwassert.FwAssert.CheckErrorMessage(
        self._worker.CallOnewayProtoRpc(
            self._start_capture_uri,
            capture_request,
            async_mode=self._async_mode))

  def stop_capture(self):
    """Concludes the current capture session."""
    capture_request = imagery_pb2.CaptureRequest()
    fwassert.FwAssert.CheckErrorMessage(
        self._worker.CallOnewayProtoRpc(
            self._stop_capture_uri,
            capture_request,
            async_mode=self._async_mode))

  def get_camera_images(self) -> Dict[Text, Sequence[Any]]:
    """Gets the latest camera images.

    Camera images can only be obtained after self.start_capture() is called.

    Returns:
      A dictionary of camera frames, with the camera id as the key. Each camera
      frame may contain multiple streams. For example, on a realsense camera we
      may have "rgb" and "depth" streams, depending on the configuration.
    """
    get_frame_request = imagery_pb2.GetFrameRequest()
    frame_collection = imagery_pb2.CameraFrameCollection()
    fwassert.FwAssert.CheckErrorMessage(
        self._worker.CallRoundtripProtoRpc(
            self._get_frame_uri,
            get_frame_request,
            frame_collection,
            self._rpc_timeout_sec,
            async_mode=self._async_mode))

    images_by_camera = {}
    for camera_frame in frame_collection.frames:
      camera_id = camera_frame.camera_id
      # In case we received multiple frames, we apppend them in the order
      # received.
      if camera_id in images_by_camera:
        images_by_camera[camera_id].append(
            self._convert_camera_frame_to_image_dict(camera_frame))
      else:
        images_by_camera[camera_id] = [
            self._convert_camera_frame_to_image_dict(camera_frame)
        ]
    return images_by_camera



import json
import numpy as np
import pybullet_utils.math_util as MathUtil


class ExpParams(object):
  RATE_KEY = 'Rate'
  INIT_ACTION_RATE_KEY = 'InitActionRate'
  NOISE_KEY = 'Noise'
  NOISE_INTERNAL_KEY = 'NoiseInternal'
  TEMP_KEY = 'Temp'

  def __init__(self):
    self.rate = 0.2
    self.init_action_rate = 0
    self.noise = 0.1
    self.noise_internal = 0
    self.temp = 0.1
    return

  def __str__(self):
    str = ''
    str += '{}: {:.2f}\n'.format(self.RATE_KEY, self.rate)
    str += '{}: {:.2f}\n'.format(self.INIT_ACTION_RATE_KEY, self.init_action_rate)
    str += '{}: {:.2f}\n'.format(self.NOISE_KEY, self.noise)
    str += '{}: {:.2f}\n'.format(self.NOISE_INTERNAL_KEY, self.noise_internal)
    str += '{}: {:.2f}\n'.format(self.TEMP_KEY, self.temp)
    return str

  def load(self, json_data):
    if (self.RATE_KEY in json_data):
      self.rate = json_data[self.RATE_KEY]

    if (self.INIT_ACTION_RATE_KEY in json_data):
      self.init_action_rate = json_data[self.INIT_ACTION_RATE_KEY]

    if (self.NOISE_KEY in json_data):
      self.noise = json_data[self.NOISE_KEY]

    if (self.NOISE_INTERNAL_KEY in json_data):
      self.noise_internal = json_data[self.NOISE_INTERNAL_KEY]

    if (self.TEMP_KEY in json_data):
      self.temp = json_data[self.TEMP_KEY]

    return

  def lerp(self, other, t):
    lerp_params = ExpParams()
    lerp_params.rate = MathUtil.lerp(self.rate, other.rate, t)
    lerp_params.init_action_rate = MathUtil.lerp(self.init_action_rate, other.init_action_rate, t)
    lerp_params.noise = MathUtil.lerp(self.noise, other.noise, t)
    lerp_params.noise_internal = MathUtil.lerp(self.noise_internal, other.noise_internal, t)
    lerp_params.temp = MathUtil.log_lerp(self.temp, other.temp, t)
    return lerp_params
